"""
PygButton v0.1.0

PygButton (pronounced "pig button") is a module that implements UI buttons for Pygame.
PygButton requires Pygame to be installed. Pygame can be downloaded from http://pygame.org
PygButton was developed by Al Sweigart (al@inventwithpython.com)
https://github.com/asweigart/pygbutton


Simplified BSD License:

Copyright 2012 Al Sweigart. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY Al Sweigart ''AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL Al Sweigart OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those of the
authors and should not be interpreted as representing official policies, either expressed
or implied, of Al Sweigart.
"""
import pygame
from pygame.locals import *
from typing import Callable

pygame.font.init()
PYGBUTTON_FONT = pygame.font.Font('freesansbold.ttf', 14)

BLACK     = (  0,   0,   0)
WHITE     = (255, 255, 255)
DARKGRAY  = ( 64,  64,  64)
GRAY      = (128, 128, 128)
LIGHTGRAY = (212, 208, 200)

class PygButton(object):
    def __init__(self, rect=None, caption='', bgcolor=LIGHTGRAY, fgcolor=BLACK, font=None, normal=None, down=None, highlight=None, callback:Callable=None):
        """Create a new button object. Parameters:
            rect - The size and position of the button as a pygame.Rect object
                or 4-tuple of integers.
            caption - The text on the button (default is blank)
            bgcolor - The background color of the button (default is a light
                gray color)
            fgcolor - The foreground color (i.e. the color of the text).
                Default is black.
            font - The pygame.font.Font object for the font of the text.
                Default is freesansbold in point 14.
            normal - A pygame.Surface object for the button's normal
                appearance.
            down - A pygame.Surface object for the button's pushed down
                appearance.
            highlight - A pygame.Surface object for the button's appearance
                when the mouse is over it.

            If the Surface objects are used, then the caption, bgcolor,
            fgcolor, and font parameters are ignored (and vice versa).
            Specifying the Surface objects lets the user use a custom image
            for the button.
            The normal, down, and highlight Surface objects must all be the
            same size as each other. Only the normal Surface object needs to
            be specified. The others, if left out, will default to the normal
            surface.
            """
        if rect is None:
            self._rect = pygame.Rect(0, 0, 30, 60)
        else:
            self._rect = pygame.Rect(rect)

        self.callback = callback
        self._caption = caption
        self._bgcolor = bgcolor
        self._fgcolor = fgcolor

        if font is None:
            self._font = PYGBUTTON_FONT
        else:
            self._font = font

        # tracks the state of the button
        self.buttonDown = False # is the button currently pushed down?
        self.mouseOverButton = False # is the mouse currently hovering over the button?
        self.lastMouseDownOverButton = False # was the last mouse down event over the mouse button? (Used to track clicks.)
        self._visible = True # is the button visible
        self.customSurfaces = False # button starts as a text button instead of having custom images for each surface

        if normal is None:
            # create the surfaces for a text button
            self.surfaceNormal = pygame.Surface(self._rect.size)
            self.surfaceDown = pygame.Surface(self._rect.size)
            self.surfaceHighlight = pygame.Surface(self._rect.size)
            self._update() # draw the initial button images
        else:
            # create the surfaces for a custom image button
            self.setSurfaces(normal, down, highlight)

    def handleEvent(self, eventObj):
        """All MOUSEMOTION, MOUSEBUTTONUP, MOUSEBUTTONDOWN event objects
        created by Pygame should be passed to this method. handleEvent() will
        detect if the event is relevant to this button and change its state.

        There are two ways that your code can respond to button-events. One is
        to inherit the PygButton class and override the mouse*() methods. The
        other is to have the caller of handleEvent() check the return value
        for the strings 'enter', 'move', 'down', 'up', 'click', or 'exit'.

        Note that mouseEnter() is always called before mouseMove(), and
        mouseMove() is always called before mouseExit(). Also, mouseUp() is
        always called before mouseClick().

        buttonDown is always True when mouseDown() is called, and always False
        when mouseUp() or mouseClick() is called. lastMouseDownOverButton is
        always False when mouseUp() or mouseClick() is called."""

        if eventObj.type not in (MOUSEMOTION, MOUSEBUTTONUP, MOUSEBUTTONDOWN) or not self._visible:
            # The button only cares bout mouse-related events (or no events, if it is invisible)
            return []

        retVal = []

        hasExited = False
        if not self.mouseOverButton and self._rect.collidepoint(eventObj.focal_point):
            # if mouse has entered the button:
            self.mouseOverButton = True
            self.mouseEnter(eventObj)
            retVal.append('enter')
        elif self.mouseOverButton and not self._rect.collidepoint(eventObj.focal_point):
            # if mouse has exited the button:
            self.mouseOverButton = False
            hasExited = True # call mouseExit() later, since we want mouseMove() to be handled before mouseExit()

        if self._rect.collidepoint(eventObj.focal_point):
            # if mouse event happened over the button:
            if eventObj.type == MOUSEMOTION:
                self.mouseMove(eventObj)
                retVal.append('move')
            elif eventObj.type == MOUSEBUTTONDOWN:
                self.buttonDown = True
                self.lastMouseDownOverButton = True
                self.mouseDown(eventObj)
                retVal.append('down')
        else:
            if eventObj.type in (MOUSEBUTTONUP, MOUSEBUTTONDOWN):
                # if an up/down happens off the button, then the next up won't cause mouseClick()
                self.lastMouseDownOverButton = False

        # mouse up is handled whether or not it was over the button
        doMouseClick = False
        if eventObj.type == MOUSEBUTTONUP:
            if self.lastMouseDownOverButton:
                doMouseClick = True
            self.lastMouseDownOverButton = False

            if self.buttonDown:
                self.buttonDown = False
                self.mouseUp(eventObj)
                retVal.append('up')

            if doMouseClick:
                self.buttonDown = False
                self.mouseClick(eventObj)
                retVal.append('click')

        if hasExited:
            self.mouseExit(eventObj)
            retVal.append('exit')

        return retVal

    def draw(self, surfaceObj):
        """Blit the current button's appearance to the surface object."""
        if self._visible:
            if self.buttonDown:
                surfaceObj.blit(self.surfaceDown, self._rect)
            elif self.mouseOverButton:
                surfaceObj.blit(self.surfaceHighlight, self._rect)
            else:
                surfaceObj.blit(self.surfaceNormal, self._rect)


    def _update(self):
        """Redraw the button's Surface object. Call this method when the button has changed appearance."""
        if self.customSurfaces:
            self.surfaceNormal    = pygame.transform.smoothscale(self.origSurfaceNormal, self._rect.size)
            self.surfaceDown      = pygame.transform.smoothscale(self.origSurfaceDown, self._rect.size)
            self.surfaceHighlight = pygame.transform.smoothscale(self.origSurfaceHighlight, self._rect.size)
            return

        w = self._rect.width # syntactic sugar
        h = self._rect.height # syntactic sugar

        # fill background color for all buttons
        self.surfaceNormal.fill(self.bgcolor)
        self.surfaceDown.fill(self.bgcolor)
        self.surfaceHighlight.fill(self.bgcolor)

        # draw caption text for all buttons
        captionSurf = self._font.render(self._caption, True, self.fgcolor, self.bgcolor)
        captionRect = captionSurf.get_rect()
        captionRect.center = int(w / 2), int(h / 2)
        self.surfaceNormal.blit(captionSurf, captionRect)
        self.surfaceDown.blit(captionSurf, captionRect)

        # draw border for normal button
        pygame.draw.rect(self.surfaceNormal, BLACK, pygame.Rect((0, 0, w, h)), 1) # black border around everything
        pygame.draw.line(self.surfaceNormal, WHITE, (1, 1), (w - 2, 1))
        pygame.draw.line(self.surfaceNormal, WHITE, (1, 1), (1, h - 2))
        pygame.draw.line(self.surfaceNormal, DARKGRAY, (1, h - 1), (w - 1, h - 1))
        pygame.draw.line(self.surfaceNormal, DARKGRAY, (w - 1, 1), (w - 1, h - 1))
        pygame.draw.line(self.surfaceNormal, GRAY, (2, h - 2), (w - 2, h - 2))
        pygame.draw.line(self.surfaceNormal, GRAY, (w - 2, 2), (w - 2, h - 2))

        # draw border for down button
        pygame.draw.rect(self.surfaceDown, BLACK, pygame.Rect((0, 0, w, h)), 1) # black border around everything
        pygame.draw.line(self.surfaceDown, WHITE, (1, 1), (w - 2, 1))
        pygame.draw.line(self.surfaceDown, WHITE, (1, 1), (1, h - 2))
        pygame.draw.line(self.surfaceDown, DARKGRAY, (1, h - 2), (1, 1))
        pygame.draw.line(self.surfaceDown, DARKGRAY, (1, 1), (w - 2, 1))
        pygame.draw.line(self.surfaceDown, GRAY, (2, h - 3), (2, 2))
        pygame.draw.line(self.surfaceDown, GRAY, (2, 2), (w - 3, 2))

        # draw border for highlight button
        self.surfaceHighlight = self.surfaceNormal


    def mouseClick(self, event):
        pass # This class is meant to be overridden.
    def mouseEnter(self, event):
        pass # This class is meant to be overridden.
    def mouseMove(self, event):
        pass # This class is meant to be overridden.
    def mouseExit(self, event):
        pass # This class is meant to be overridden.
    def mouseDown(self, event):
        pass # This class is meant to be overridden.
    def mouseUp(self, event):
        pass # This class is meant to be overridden.


    def setSurfaces(self, normalSurface, downSurface=None, highlightSurface=None):
        """Switch the button to a custom image type of button (rather than a
        text button). You can specify either a pygame.Surface object or a
        string of a filename to load for each of the three button appearance
        states."""
        if downSurface is None:
            downSurface = normalSurface
        if highlightSurface is None:
            highlightSurface = normalSurface

        if type(normalSurface) == str:
            self.origSurfaceNormal = pygame.image.load(normalSurface)
        if type(downSurface) == str:
            self.origSurfaceDown = pygame.image.load(downSurface)
        if type(highlightSurface) == str:
            self.origSurfaceHighlight = pygame.image.load(highlightSurface)

        if self.origSurfaceNormal.get_size() != self.origSurfaceDown.get_size() != self.origSurfaceHighlight.get_size():
            raise Exception('foo')

        self.surfaceNormal = self.origSurfaceNormal
        self.surfaceDown = self.origSurfaceDown
        self.surfaceHighlight = self.origSurfaceHighlight
        self.customSurfaces = True
        self._rect = pygame.Rect((self._rect.left, self._rect.top, self.surfaceNormal.get_width(), self.surfaceNormal.get_height()))



    def _propGetCaption(self):
        return self._caption


    def _propSetCaption(self, captionText):
        self.customSurfaces = False
        self._caption = captionText
        self._update()


    def _propGetRect(self):
        return self._rect


    def _propSetRect(self, newRect):
        # Note that changing the attributes of the Rect won't update the button. You have to re-assign the rect member.
        self._update()
        self._rect = newRect


    def _propGetVisible(self):
        return self._visible


    def _propSetVisible(self, setting):
        self._visible = setting


    def _propGetFgColor(self):
        return self._fgcolor


    def _propSetFgColor(self, setting):
        self.customSurfaces = False
        self._fgcolor = setting
        self._update()


    def _propGetBgColor(self):
        return self._bgcolor


    def _propSetBgColor(self, setting):
        self.customSurfaces = False
        self._bgcolor = setting
        self._update()


    def _propGetFont(self):
        return self._font


    def _propSetFont(self, setting):
        self.customSurfaces = False
        self._font = setting
        self._update()


    caption = property(_propGetCaption, _propSetCaption)
    rect = property(_propGetRect, _propSetRect)
    visible = property(_propGetVisible, _propSetVisible)
    fgcolor = property(_propGetFgColor, _propSetFgColor)
    bgcolor = property(_propGetBgColor, _propSetBgColor)
    font = property(_propGetFont, _propSetFont)


from coopgame.coordinateConverter import DrawMatrixState, MultiPerspectivePointConverter
from coopgame.renderedObjectHandling.wireframes.WireframeObjects import Cube
import pygame
from coopgame.gameTemplate import GameTemplate
from coopstructs.geometry.vectors.vectorN import Vector3, Vector2
from cooptools.colors import Color
from coopstructs.geometry import Rectangle
from coopgame.renderedObjectHandling.renderedViewArgs import RenderedViewArgs
from coopgame.renderedObjectHandling.wireframes.Wireframe import Wireframe
from coopgame.renderedObjectHandling.renderedObject import RenderedObject
import coopgame.pygamehelpers as help
import coopgame.label_drawing.label_drawing_utils as lab
from coopgame.handleKeyPressedArgs import InputState, InputStateHandler, InputAction, InputEvent, InputEventType, basic_key_actions, CallbackPackage
from coopgame.pygame_k_constant_names import PyKeys, PyMouse

FOCUSED_OBJ = 'cube'
class CoordinateConversionGame(GameTemplate):

    def __init__(self):
        super().__init__(fullscreen=False, input_state_handler=self.define_input_handler())
        pygame.display.set_caption('Wireframe Display')

        self.mppc = MultiPerspectivePointConverter()
        self.matrix_state = DrawMatrixState()

        self.cube = Cube((0, 0, 0), 1)
        self.rendered_objects = {}
        self.focused_object = FOCUSED_OBJ


        self.viewargs = RenderedViewArgs()
        self.radians_per_sec = .1

        self.displayNodes = True
        self.displayEdges = True
        self.displayCenters = True
        self.nodeColour = Color.ORANGE
        self.edgeColour = Color.MAGENTA
        self.centerColour = Color.RED
        self.nodeRadius = 2
        self.centerNodeRadius = 4

    def define_input_handler(self):
        return InputStateHandler(
            quit_callback_package=CallbackPackage(down=lambda x: self.quit()),
            debug_callback_package=CallbackPackage(down=lambda x: self.debug_mode_toggle.toggle()),
            fullscreen_callback_package=CallbackPackage(down=lambda x: self.toggle_fullscreen()),
            key_actions=[
                # OBJ MOVE
                InputAction([InputEvent(PyKeys.K_w, event_type=InputEventType.HELD)], callback=lambda x: self.rendered_objects[self.focused_object].orientation.change_translation(delta_vector=Vector3(0, -1, 0))),
                InputAction([InputEvent(PyKeys.K_a, event_type=InputEventType.HELD)], callback=lambda x: self.rendered_objects[self.focused_object].orientation.change_translation(delta_vector=Vector3(-1, 0, 0))),
                InputAction([InputEvent(PyKeys.K_s, event_type=InputEventType.HELD)], callback=lambda x: self.rendered_objects[self.focused_object].orientation.change_translation(delta_vector=Vector3(0, 1, 0))),
                InputAction([InputEvent(PyKeys.K_d, event_type=InputEventType.HELD)], callback=lambda x: self.rendered_objects[self.focused_object].orientation.change_translation(delta_vector=Vector3(1, 0, 0))),

                #OBJ ROT
                InputAction([InputEvent(PyKeys.K_q, event_type=InputEventType.HELD)], callback=lambda x: self.rendered_objects[self.focused_object].orientation.change_rotation(rads_delta=.1, rotation_axis=Vector3(0, 1, 0))),
                InputAction([InputEvent(PyKeys.K_e, event_type=InputEventType.HELD)], callback=lambda x: self.rendered_objects[self.focused_object].orientation.change_rotation(rads_delta=-.1, rotation_axis=Vector3(0, 1, 0))),

                # CAMERA MOVE
                InputAction([InputEvent(PyKeys.K_LEFT, event_type=InputEventType.HELD)], callback=lambda x: self.viewargs.camera_orientation.change_translation(delta_vector=Vector3(-1, 0, 0))),
                InputAction([InputEvent(PyKeys.K_RIGHT, event_type=InputEventType.HELD)], callback=lambda x: self.viewargs.camera_orientation.change_translation(delta_vector=Vector3(1, 0, 0))),
                InputAction([InputEvent(PyKeys.K_UP, event_type=InputEventType.HELD)], callback=lambda x: self.viewargs.camera_orientation.change_translation(delta_vector=Vector3(0, -1, 0))),
                InputAction([InputEvent(PyKeys.K_DOWN, event_type=InputEventType.HELD)], callback=lambda x: self.viewargs.camera_orientation.change_translation(delta_vector=Vector3(0, 1, 0))),

                # CAMERA ROT
                InputAction([InputEvent(PyKeys.K_i, event_type=InputEventType.HELD)], callback=lambda x: self.viewargs.camera_orientation.change_rotation(rads_delta=.1, rotation_axis=Vector3(1, 0, 0))),
                InputAction([InputEvent(PyKeys.K_k, event_type=InputEventType.HELD)], callback=lambda x: self.viewargs.camera_orientation.change_rotation(rads_delta=-.1, rotation_axis=Vector3(1, 0, 0))),
                InputAction([InputEvent(PyKeys.K_j, event_type=InputEventType.HELD)], callback=lambda x: self.viewargs.camera_orientation.change_rotation(rads_delta=.1, rotation_axis=Vector3(0, 1, 0))),
                InputAction([InputEvent(PyKeys.K_l, event_type=InputEventType.HELD)], callback=lambda x: self.viewargs.camera_orientation.change_rotation(rads_delta=-.1, rotation_axis=Vector3(0, 1, 0))),

        ])

    def initialize_game(self):
        pass

    def handle_hover_over(self, mouse_pos_as_vector: Vector2):
        pass

    def handle_left_mouse_down(self, pressed_keys):
        pass

    def handle_right_mouse_down(self, pressed_keys):
        pass

    def handle_mouse_scroll_up(self, pressed_keys):
        self.viewargs.camera_orientation.change_scale(Vector3(1.1, 1.1, 1.1), Vector3(*self.rendered_objects[self.focused_object].camera_center(self.viewargs)[:3]))

    def handle_mouse_scroll_down(self, pressed_keys):
        self.viewargs.camera_orientation.change_scale(Vector3(.9, .9, .9), Vector3(*self.rendered_objects[self.focused_object].camera_center(self.viewargs)[:3]))

    def model_updater(self, delta_time_ms: int):
        pass
        #     self.matrix_state.set_matricies(self.viewargs)
        #     game_area_rect = Rectangle(0, 0, self.screen.get_height(), self.screen.get_width())
        #     self.viewport_points = self.mppc.convert_real_to_viewport([Vector3(point[0], point[1], point[2]) for point in self.cube.nodes],
        #                                                               game_area_rect=game_area_rect,
        #                                                               matrix_state=self.matrix_state)
    def addRenderedObject(self,
                          name: str,
                          wireframe: Wireframe):
        """ Add a named wireframe object. """
        self.rendered_objects[name] = RenderedObject(wireframe)

    def draw(self, frames:int, debug_mode: bool = False):
        self._draw_rendered_objects()
        lab.draw_label(
            hud=self.screen,
            text=f"{self.rendered_objects[self.focused_object].world_center[:3]}",
            args=lab.DrawLabelArgs(),
            pos=(0, self.screen.get_height() - 100)
        )

        lab.draw_label(
            hud=self.screen,
            text=f"{self.rendered_objects[self.focused_object].camera_center(self.viewargs)[:3]}",
            args=lab.DrawLabelArgs(),
            pos=(0, self.screen.get_height() - 50)
        )

        lab.draw_label(
            hud=self.screen,
            text=str(help.mouse_pos()),
            args=lab.DrawLabelArgs(),
            pos=(0, self.screen.get_height())
        )





        # help.draw_text(text=f"{self.rendered_objects[self.focused_object].world_center[:3]}",
        #                hud=self.screen,
        #                offset_rect=Rectangle(0, self.screen.get_height() - 100, 20, self.screen.get_width()))
        # help.draw_text(text=f"{self.rendered_objects[self.focused_object].camera_center(self.viewargs)[:3]}",
        #                hud=self.screen)
        #
        # help.draw_mouse_coord(self.screen,
        #                       offset_rect=Rectangle(0, self.screen.get_height() - 150, 20, self.screen.get_width()))

    def _draw_rendered_objects(self):
        for name, obj in self.rendered_objects.items():
            if self.displayEdges:
                for n1, n2 in obj.camera_edges(self.viewargs):
                    pygame.draw.aaline(self.screen, self.edgeColour.value, n1, n2, 1)

            if self.displayNodes:
                for node in obj.camera_nodes(self.viewargs):
                    pygame.draw.circle(self.screen, self.nodeColour.value, (int(node[0]), int(node[1])), self.nodeRadius, 0)

            if self.displayCenters:
                center = obj.camera_center(self.viewargs)
                pygame.draw.circle(self.screen, self.centerColour.value, (int(center[0]), int(center[1])),
                                   self.centerNodeRadius, 0)


    def sprite_updater(self, delta_time_ms: int):
        pass

    def on_resize(self):
        pass


if __name__ == "__main__":

    cube = Cube((50, 50, 50), 100)


    pv = CoordinateConversionGame()
    pv.addRenderedObject(FOCUSED_OBJ, cube)

    # axis = np.array([1, 0, 0])
    # rotation_vector = axis / (axis ** 2).sum() ** 0.5
    #
    # pv._rotate_all(rotation_vector, np.pi / 3)

    pv.main()


from cooptools.colors import Color
import pygame
from cooptools.graphs import Graph, AStarResults, Node, Edge
import coopgame.pygamehelpers as help
from typing import List, Callable, Iterable, Tuple
import coopgame.grids_graphs.draw_graph_utils as utils
import coopgame.linedrawing.line_draw_utils as lutils
import coopgame.label_drawing.label_drawing_utils as labutils
from cooptools.coopEnum import CoopEnum
from coopgame.surfaceManager import SurfaceManager, SurfaceRegistryArgs, SurfaceGroup
import coopgame.pointdrawing.point_draw_utils as putils
import cooptools.geometry_utils.vector_utils as vec
import logging

logger = logging.getLogger(__name__)

class GraphSurfaceType(CoopEnum):
    EDGES_SURFACE_ID = 'EDGES_SURFACE_ID'
    NODES_SURFACE_ID = 'NODES_SURFACE_ID'
    NODE_LABELS_SURFACE_ID = 'NODE_LABELS_SURFACE_ID'
    EDGE_LABELS_SURFACE_ID = 'EDGE_LABELS_SURFACE_ID'
    OVERLAY_SURFACE_ID = 'OVERLAY_SURFACE_ID'
    ROUTES_SURFACE_ID = 'ROUTES_SURFACE_ID'
    NODE_LABELS_SURFACE_GROUP_ID = 'NODE_LABELS_SURFACE_GROUP_ID'
    EDGE_LABELS_SURFACE_GROUP_ID = 'EDGE_LABELS_SURFACE_GROUP_ID'

DEFAULT_DRAW_CONFIG = utils.GraphDrawArgs(
            node_draw_args=putils.DrawPointArgs(
                color=Color.DARK_BLUE,
                radius=5,
            ),
            enabled_edge_args=lutils.DrawLineArgs(
                color=Color.ORANGE,
                directionality_draw_args=lutils.DirectionalityIndicatorDrawArgs(
                    color=Color.ORANGE,
                    height=10,
                    width=10
                ),
                draw_label_args=labutils.DrawLabelArgs(
                    color=Color.WHEAT
                )
            ),
            disabled_edge_args=lutils.DrawLineArgs(
                color=Color.BROWN,
                directionality_draw_args=lutils.DirectionalityIndicatorDrawArgs(
                    color=Color.BROWN,
                    height=10,
                    width=10
                ),
                draw_label_args=labutils.DrawLabelArgs(
                    color=Color.BROWN
                )
            ),
            node_label_args=labutils.DrawLabelArgs(
                color=Color.WHEAT
            ),
            # articulation_points_args=putils.DrawPointArgs(
            #     outline_color=Color.PURPLE,
            #     radius=10,
            #     outline_width=3
            # ),
            sink_node_args=putils.DrawPointArgs(
                color=Color.RED,
                radius=10
            ),
            source_node_args=putils.DrawPointArgs(
                color=Color.GREEN,
                radius=10
            ),
            orphan_node_args=putils.DrawPointArgs(
                outline_color=Color.RED,
                radius=10
            )
        )

GraphGetter = Callable[[], Graph]
RouteDrawArgsGetter = Callable[[], List[Tuple[AStarResults, utils.RouteDrawArgs]]]

class PyGraphHandler:
    def __init__(self,
                 screen: pygame.Surface,
                 graph_getter: GraphGetter,
                 draw_config: utils.GraphDrawArgs = None,
                 route_draw_args_getter: RouteDrawArgsGetter = None,
                 vec_transformer: vec.VecTransformer = None):
        self.parent_screen = screen
        self.graph_getter = graph_getter
        self._draw_config: utils.GraphDrawArgs = draw_config
        self._route_draw_args_getter = route_draw_args_getter
        self._vec_transformer = vec_transformer

        self.surface_manager = SurfaceManager(
            surface_draw_callbacks=[
                SurfaceRegistryArgs(GraphSurfaceType.EDGES_SURFACE_ID.value, self.redraw_edges_surface),
                SurfaceRegistryArgs(GraphSurfaceType.NODES_SURFACE_ID.value, self.redraw_nodes_surface),
                SurfaceRegistryArgs(GraphSurfaceType.EDGE_LABELS_SURFACE_ID.value, self.redraw_edge_labels_surface, default_visible=False),
                SurfaceRegistryArgs(GraphSurfaceType.NODE_LABELS_SURFACE_ID.value, self.redraw_node_labels_surface, default_visible=False),
                SurfaceRegistryArgs(GraphSurfaceType.OVERLAY_SURFACE_ID.value, self.redraw_overlay_surface),
                SurfaceRegistryArgs(GraphSurfaceType.ROUTES_SURFACE_ID.value, self.redraw_routes_surface),
                SurfaceRegistryArgs(GraphSurfaceType.NODE_LABELS_SURFACE_GROUP_ID.value, self.redraw_node_labels_surface_group),
                SurfaceRegistryArgs(GraphSurfaceType.EDGE_LABELS_SURFACE_GROUP_ID.value, self.redraw_edge_labels_surface_group),
            ]
        )

        self._surface_group_id_whitelist = {

        }

    def set_config(self,
                   draw_config: utils.GraphDrawArgs):
        self._draw_config = draw_config

    def redraw_edges_surface(self):
        if self._draw_config is None:
            return None
        surf = help.init_surface(self.parent_screen.get_size())
        utils.draw_to_surface(
            surface=surf,
            graph=self.graph_getter(),
            args=self._draw_config.EdgesBaseArgs,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_nodes_surface(self):
        if self._draw_config is None:
            return None

        surf = help.init_surface(self.parent_screen.get_size())
        utils.draw_to_surface(
            surface=surf,
            graph=self.graph_getter(),
            args=self._draw_config.NodesBaseArgs,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_node_labels_surface(self):
        if self._draw_config is None:
            return None

        surf = help.init_surface(self.parent_screen.get_size())
        utils.draw_to_surface(
            surface=surf,
            graph=self.graph_getter(),
            args=self._draw_config.NodesLabelArgs,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_edge_labels_surface(self):
        if self._draw_config is None:
            return None

        surf = help.init_surface(self.parent_screen.get_size())
        utils.draw_to_surface(
            surface=surf,
            graph=self.graph_getter(),
            args=self._draw_config.EdgesLabelArgs,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_overlay_surface(self):
        if self._draw_config is None:
            return None

        surf = help.init_surface(self.parent_screen.get_size())
        utils.draw_to_surface(
            surface=surf,
            graph=self.graph_getter(),
            args=self._draw_config.OverlayArgs,
            vec_transformer=self._vec_transformer
        )
        return surf

    def redraw_routes_surface(self):
        surf = help.init_surface(self.parent_screen.get_size())

        if self._route_draw_args_getter is not None:
            utils.draw_to_surface(
                surface=surf,
                graph=self.graph_getter(),
                routes=self._route_draw_args_getter(),
                vec_transformer=self._vec_transformer
            )

        return surf

    def redraw_node_labels_surface_group(self):
        if self._draw_config is None:
            return None

        return utils.get_node_label_surface_group(
            graph=self.graph_getter(),
            args=self._draw_config,
            vec_transformer=self._vec_transformer
        )

    def redraw_edge_labels_surface_group(self):
        if self._draw_config is None:
            return None

        return utils.get_edge_label_surface_group(
            graph=self.graph_getter(),
            args=self._draw_config,
            vec_transformer=self._vec_transformer
        )

    def invalidate(self, surfaces: Iterable[GraphSurfaceType] = None):
        surf_names = [x.name for x in surfaces] if surfaces is not None else None
        self.surface_manager.invalidate(surf_names)

    def redraw(self):
        self.surface_manager.redraw([x.id for x in GraphSurfaceType])

    def render(self,
               surface: pygame.Surface):
        self.update()
        self.surface_manager.render(surface,
                                    surface_group_ids_whitelist=self._surface_group_id_whitelist)

    def toggle_surface(self, graphSurfaceTypes: List[GraphSurfaceType]):
        self.surface_manager.toggle_visible([x.value for x in graphSurfaceTypes])

    def show_all(self):
        self.surface_manager.show_all()

    def hide_all(self):
        self.surface_manager.hide_all()


    def _naive_mouse_close_to_nodes(self,
                                    graph: Graph,
                                    mouse_pos: Iterable[float],
                                    threshold: float
                                    ) -> List[Node]:
        ret = []
        for id, node in graph.Nodes.items():
            if vec.distance_between(a=mouse_pos,
                                    b=vec.resolve_vec_transformed_points([node.pos], self._vec_transformer)[0],
                                    allow_diff_lengths=True) <= threshold:
                ret.append(node)

        return ret

    def _naive_mouse_close_to_edges(self,
                                    graph: Graph,
                                    mouse_pos: Iterable[float],
                                    threshold: float
                                    ) -> List[Edge]:
        ret = []
        for edge in graph.Edges.values():
            end = vec.resolve_vec_transformed_points([edge.end.pos], self._vec_transformer)[0]
            start = vec.resolve_vec_transformed_points([edge.start.pos], self._vec_transformer)[0]
            projection = vec.project_onto(a=mouse_pos,
                                          b=end,
                                          origin=start,
                                          allow_diff_lengths=True)
            if vec.distance_between(mouse_pos, projection, allow_diff_lengths=True) <= threshold\
                    and vec.bounded_by(projection, start, end):
                ret.append(edge)

        return ret

    def handle_hover(self):
        close_nodes = self._naive_mouse_close_to_nodes(
            graph=self.graph_getter(),
            mouse_pos=help.mouse_pos(),
            threshold=15
        )

        close_edges = self._naive_mouse_close_to_edges(
            graph=self.graph_getter(),
            mouse_pos=help.mouse_pos(),
            threshold=15
        )

        logger.debug(f"CloseNodes: {close_nodes}")
        logger.debug(f"CloseEdges: {close_edges}")

        self._surface_group_id_whitelist = {
            GraphSurfaceType.NODE_LABELS_SURFACE_GROUP_ID.value: [x.name for x in close_nodes],
            GraphSurfaceType.EDGE_LABELS_SURFACE_GROUP_ID.value: [x.id for x in close_edges]
        }

    def update(self):
        self.handle_hover()




import cooptools.sectors.rect_utils as sec_utl
from typing import Tuple

class ScrollHandler:
    def __init__(self,
                 screen_dims: Tuple[int, int],
                 focal_point: Tuple[float, float] = None,
                 width_dim_p: Tuple[float, float] = None,
                 height_dim_p: Tuple[float, float] = None):
        self.scroll = (0, 0)
        self.width_dim_p = width_dim_p or (0.25, 0.75)
        self.height_dim_p = height_dim_p or (0.25, 0.75)
        self.screen_dims = screen_dims
        self.focal_point = None

        if focal_point:
            self.update(focal_point)

    def update(self,
               focal_point: Tuple[float, float],
               screen_dims: Tuple[int, int] = None,
               width_dim_p: Tuple[float, float] = None,
               height_dim_p: Tuple[float, float] = None
               ):
        if width_dim_p is not None:
            self.width_dim_p = width_dim_p

        if height_dim_p is not None:
            self.height_dim_p = height_dim_p

        if screen_dims is not None:
            self.screen_dims = screen_dims

        delta_x = 0
        delta_y = 0

        self.focal_point = focal_point

        ciw = self.WindowCoord
        if ciw[0] > self.screen_dims[0] * self.width_dim_p[1]:
            delta_x = ciw[0] - self.screen_dims[0] * self.width_dim_p[1]

        if ciw[0] < self.screen_dims[0] * self.width_dim_p[0]:
            delta_x = ciw[0] - self.screen_dims[0] * self.width_dim_p[0]

        if ciw[1] > self.screen_dims[1] * self.height_dim_p[1]:
            delta_y = ciw[1] - self.screen_dims[1] * self.height_dim_p[1]

        if ciw[1] < self.screen_dims[1] * self.height_dim_p[0]:
            delta_y = ciw[1] - self.screen_dims[1] * self.height_dim_p[0]

        self.scroll = (self.scroll[0] + delta_x, self.scroll[1] + delta_y)
        return self.scroll

    def coord_in_window(self, coord_actual: Tuple[float, float]):
        return (coord_actual[0] - self.scroll[0], coord_actual[1] - self.scroll[1])

    @property
    def WindowCoord(self) -> Tuple[float, float]:
        return self.coord_in_window(self.focal_point)


class ScrollTileHandler:

    def __init__(self,
                 tile_size: Tuple[float, float],
                 screen_dims: Tuple[int, int],
                 focal_point: Tuple[float, float],
                 width_dim_p: Tuple[float, float] = None,
                 height_dim_p: Tuple[float, float] = None
                 ):
        self.tile_size = tile_size
        self.scroll_handler = ScrollHandler(screen_dims=screen_dims,
                                            focal_point=focal_point,
                                            width_dim_p=width_dim_p,
                                            height_dim_p=height_dim_p)

    def update(self,
               focal_point: Tuple[float, float],
               screen_dims: Tuple[int, int] = None,
               width_dim_p: Tuple[float, float] = None,
               height_dim_p: Tuple[float, float] = None
               ):
        return self.scroll_handler.update(
            focal_point=focal_point,
            screen_dims=screen_dims,
            width_dim_p=width_dim_p,
            height_dim_p=height_dim_p
        )

    def tile_window_coord(self, tile):
        tile_coords = sec_utl.coord_of_sector(
            sector_dims = self.tile_size,
            sector = tile
        )
        return self.scroll_handler.coord_in_window(tile_coords)


    @property
    def ScrollTile(self):
        """What tile is the current scroll value in?"""
        return sec_utl.sector_from_coord(
            coord=self.scroll_handler.scroll,
            sec_dims=self.tile_size)

    @property
    def ScrollTileCoords(self):
        """What are the coords of the tile that we are scrolled in"""
        return sec_utl.coord_of_sector(
            sector_dims=self.tile_size,
            sector=self.ScrollTile
        )

    @property
    def WorldTile(self):
        return sec_utl.sector_from_coord(
            coord=self.scroll_handler.focal_point,
            sec_dims=self.tile_size
        )

    @property
    def WorldTileCoord(self):
        return sec_utl.coord_of_sector(
            sector=self.WorldTile,
            sector_dims=self.tile_size
        )

    @property
    def WindowTile(self):
        return sec_utl.sector_from_coord(
            coord = (self.scroll_handler.focal_point[0] + self.scroll_handler.scroll[0],
                    self.scroll_handler.focal_point[1] + self.scroll_handler.scroll[1]),
            sec_dims=self.tile_size
        )

    @property
    def WindowTileCoord(self):
        return sec_utl.coord_of_sector(
            sector=self.WindowTile,
            sector_dims=self.tile_size
        )

    @property
    def PosInCurrentTile(self):
        world_tile_coord = self.WorldTileCoord
        return (self.scroll_handler.focal_point[0] - world_tile_coord[0],
                self.scroll_handler.focal_point[1] - world_tile_coord[1])

    @property
    def WorldCoordOffsetScrollTileCoord(self):
        return (self.scroll_handler.focal_point[0] - self.ScrollTileCoords[0] - self.scroll_handler.WindowCoord[0],
                self.scroll_handler.focal_point[1] - self.ScrollTileCoords[1] - self.scroll_handler.WindowCoord[1])

    @property
    def WindowCoords(self):
        return self.scroll_handler.WindowCoord

    @property
    def Scroll(self):
        return self.scroll_handler.scroll




from dataclasses import dataclass, field, asdict
from typing import List, Dict, Iterable, Callable
import uuid
import coopgame.characterDevelopment.stats as stat
from cooptools.transform import Transform
from cooptools.typeProviders import StringProvider, resolve_string_provider

NameProvider = Callable[[str], str] | StringProvider

@dataclass(frozen=True, slots=True)
class CharacterGeneratorArgs:
    name_provider: NameProvider
    init_stats_package: stat.StatsPackageState
    traits_manifest: stat.StatEffectPackageManifest

@dataclass(frozen=True, slots=True)
class Opinion:
    factors: Dict[str, float]

    @property
    def Score(self):
        return sum(self.factors.values())


class Character:
    def __init__(self,
                name_provider: NameProvider,
                init_stats: stat.StatsPackageState,
                traits: Iterable[stat.StatEffectPackage]=None):
        self.id = str(uuid.uuid4())
        self.stats: stat.StatsPackageCommandController=stat.StatsPackageCommandController()\
            .init_stats(stat_args=init_stats.base_stats.stats.values())
        self.transform: Transform=Transform()
        self.stats.acquire_stat_effect_packages(stat_effect_packages=traits)
        self.name: str=self._resolve_name(name_provider)

    def __repr__(self):
        return f"{self.name}, [{self.id}] ({', '.join([x.id for x in self.stats.State.stat_modifiers])})"

    @classmethod
    def generate(self,
                 args: CharacterGeneratorArgs,
                 trait_count_args: Dict[str, int]):
        traits = []

        for trait_type, count in trait_count_args.items():
            used = []
            for ii in range(count):
                type, choice = args.traits_manifest.stat_effect_type_manifests[trait_type].random(excluded_types=used)
                used.append(type)
                traits.append(choice)

        return Character(
            name_provider=args.name_provider,
            init_stats=args.init_stats_package,
            traits=traits
        )

    def _resolve_name(self, name_provider: NameProvider):
        try:
            return name_provider(self.Race.id)
        except:
            return resolve_string_provider(name_provider)

    def opinon_of(self, character) -> Opinion:
        if type(character) != Character:
            raise TypeError(f"Cannot calculate opinion of type {type(character)}")

        factors = {}
        for stat_package in self.stats.State.stat_modifiers:
            if stat_package.stat_opinions is None:
                deb = True
            for trait, effect in stat_package.stat_opinions.items():
                if trait in character.stats.State.stat_modifiers:
                    factors[trait] = effect

        return Opinion(
            factors=factors
        )

    @property
    def Race(self):
        ret = next((i for i in self.stats.State.stat_modifiers if i.category == 'race'), None)
        return ret

if __name__ == "__main__":
    from pprint import pprint

    config_fn_provider = r'C:\Users\tburns\PycharmProjects\coopgame\coopgame\characterDevelopment\manifest.json'
    me = Character.generate(
        args=CharacterGeneratorArgs(
            name_provider="Ralphius",
            init_stats_package=stat.StatsPackageState.from_config(config_fn_provider),
            traits_manifest=stat.StatEffectPackageManifest.from_config_data(config_fn_provider)
        ),
        trait_count_args={
            "personality_traits": 3,
            "physical_traits": 2,
            "race": 1
        }
    )

    # pprint(me.stats.State.Summary)
    pprint(me)
    me.stats.register_events(events=[
        stat.DeltaStatStatsPackageStateEvent(stat_name='HP', delta=-19),
        stat.StatsPackageStateEvent(description="character gained rage",
                                    added_stat_modifiers=[
                            stat.StatEffectPackage(
                                stat_effects=frozenset([
                                    stat.StatEffect(stat_name='rage', delta_effect=5)]),
                                id='Enraged'
                            )
                                    ],

                                    )
    ])

    pprint(me)
    pprint(me.stats.State.Summary)

    print(f"All My Events:")
    pprint(me.stats.Events)


import pygame
from coopgame.gameTemplate import GameTemplate
from coopgame.tilemap import Tilemap, FileTileset
import numpy as np
from cooptools.colors import Color
from coopgame.handleKeyPressedArgs import InputState, InputStateHandler, InputAction, InputEvent, InputEventType, basic_key_actions, CallbackPackage
from coopgame.pygame_k_constant_names import PyKeys, PyMouse
import coopgame.pygamehelpers as phelp
import cooptools.sectors.rect_utils as rutils

file = '../coopgame/assets/tileSets/tmw_desert_spacing.png'

class TileGame(GameTemplate):
    W = 640
    H = 240
    SIZE = W, H

    def __init__(self):


        root_keys = [
            InputAction([InputEvent(ii, event_type=InputEventType.DOWN)], callback=lambda x, d=ii: self.set_value_at_mouse_pos(d.value-48))
            for ii in [PyKeys.K_0, PyKeys.K_1, PyKeys.K_2, PyKeys.K_3, PyKeys.K_4, PyKeys.K_5, PyKeys.K_6, PyKeys.K_7, PyKeys.K_8, PyKeys.K_9]
        ]

        input_state_handler = InputStateHandler(
            quit_callback_package=CallbackPackage(down=lambda x: self.quit()),
            debug_callback_package=CallbackPackage(down=lambda x: self.debug_mode_toggle.toggle()),
            fullscreen_callback_package=CallbackPackage(down=lambda x: self.toggle_fullscreen()),
            key_actions=root_keys + [
                InputAction([InputEvent(PyKeys.K_r, event_type=InputEventType.DOWN)], callback=lambda x: self.tilemap.set_random()),
                InputAction([InputEvent(PyKeys.K_l, event_type=InputEventType.DOWN)], callback=lambda x: self.load_image(file)),
                InputAction([InputEvent(PyMouse.LEFT, event_type=InputEventType.DOWN)], callback=lambda x: self.increment_value_at_mouse_pos(increment=1)),
                InputAction([InputEvent(PyMouse.RIGHT, event_type=InputEventType.DOWN)], callback=lambda x: self.increment_value_at_mouse_pos(increment=-1))
        ])

        super().__init__(input_state_handler=input_state_handler)
        self.screen = pygame.display.set_mode(TileGame.SIZE)
        self.tileset = FileTileset(file)
        self.tilemap = Tilemap(self.tileset,
                               map=np.array([[0, 6, 12],
                                             [1, 7, 13],
                                             [2, 8, 14]]),
                               surface_size=TileGame.SIZE)
        pygame.display.set_caption("Pygame Tiled Demo")
        self.image = None


    def draw(self, frames: int, debug_mode: bool = False):
        self.tilemap.draw(self.screen, outline_tiles_color=Color.RED)

    def load_image(self, file):
        try:
            self.file = file
            self.image = pygame.image.load(file)
            self.rect = self.image.get_rect()
            self.screen = pygame.display.set_mode(self.rect.size)
            pygame.display.set_caption(f'size:{self.rect.size}')

        except Exception as e:
            raise Exception(f"Unable to load file: {file}: {e}")

    def set_value_at_mouse_pos(self, value):
        mouse_grid_pos = rutils.sector_from_coord(
            coord=phelp.mouse_pos(),
            sec_dims=self.tilemap.TileSize
        )
        self.tilemap.set_value(value, positions=[mouse_grid_pos])

    def increment_value_at_mouse_pos(self, increment: int = 1):
        mouse_grid_pos = rutils.sector_from_coord(
            coord=phelp.mouse_pos(),
            sec_dims=self.tilemap.TileSize
        )
        self.tilemap.increment_value(positions=[mouse_grid_pos], increment=increment)

if __name__ == "__main__":

    game = TileGame()
    game.main()


from coopgame.enums import Rarity
class ItemBase:
    def __init__(self,
                 rarity: Rarity):
        self._rarity: Rarity = rarity

    @property
    def Rarity(self) -> Rarity:
        return self._rarity



from coopgame.particleSystem.emitter import Emitter
from coopgame.particleSystem.destroyers import time_destroyer, time_and_size_destroyer, CompareArgs
from coopstructs.geometry.vectors.vectorN import Vector2
from functools import partial
from typing import Union, Callable, Tuple
import coopgame.particleSystem.utils as utils
from coopgame.particleSystem.callables import vec2_callable, int_callable, float_callable, phase_callback
import random as rnd
from coopgame.particleSystem.rendering.particleRenderer import ParticleRenderer
from coopgame.particleSystem.emitterComposite import EmitterComposite
from coopgame.particleSystem.particleGenerationArgs import ParticleGenerationArgs, BurstArgs, PhaseArgs
from coopgame.particleSystem.emitterArgs import EmitterArgs
from coopgame.particleSystem.protocol import UpdateRenderableProtocol
from coopgame.particleSystem.emitterPhaseController import EmitterPhaseController
from coopgame.particleSystem.particleGenerationArgs import EmitterPhase


def snow(origin_line_start: Vector2,
         origin_line_end: Vector2,
         fall_velo: Vector2,
         lifespan_ms: Union[int, Callable],
         size: float,
         renderer: ParticleRenderer,
         twinkle_ms: float,
         particle_generation_args: ParticleGenerationArgs = None) -> UpdateRenderableProtocol:
    from_top_origin = lambda: utils.rand_between_points(origin_line_start, origin_line_end)
    return Emitter(EmitterArgs(
        origin=from_top_origin,
        velocity=fall_velo,
        destroyers=[partial(time_destroyer, lifespan_ms=lifespan_ms)],
        size=lambda x: max(utils.sin_wave(x, utils.SinWaveArgs(amp=size, wavelength=twinkle_ms, norm_positive=True)), 1),
        renderer=renderer,
        particle_generation_args=particle_generation_args),
        id='snow'
    )

def comet(path: vec2_callable,
          radiate_speed: int,
          lifespan_ms: int,
          start_size: float,
          renderer: ParticleRenderer,
          particle_generation_args: ParticleGenerationArgs = None) -> UpdateRenderableProtocol:
    radiate_velo = lambda: Vector2(rnd.random() * 2 - 1, rnd.random() * 2 - 1) * radiate_speed

    return Emitter(EmitterArgs(
        origin=path,
        velocity=radiate_velo,
        destroyers=[partial(time_destroyer, lifespan_ms=lifespan_ms)],
        size=lambda x: utils.decay(x, lifespan_ms=lifespan_ms, max_size=start_size),
        renderer=renderer,
        particle_generation_args=particle_generation_args),
        id='comet'
    )


def torch(origin: Union[Vector2, vec2_callable],
          fire_lifespan_ms: int,
          smoke_lifespan_ms: int,
          fire_start_size: float,
          smoke_end_size: float,
          fire_raise_velo: Union[Vector2, vec2_callable],
          smoke_raise_velo: Union[Vector2, vec2_callable],
          fire_renderer: ParticleRenderer,
          smoke_renderer: ParticleRenderer,
          fire_generation_args: ParticleGenerationArgs = None,
          smoke_generation_args: ParticleGenerationArgs = None) -> UpdateRenderableProtocol:

    if fire_generation_args is None:
        fire_generation_args = ParticleGenerationArgs(generation_interval_ms=10)
    if smoke_generation_args is None:
        smoke_generation_args = ParticleGenerationArgs(generation_interval_ms=500)

    if not callable(origin):
        origin = lambda: origin

    # if not callable(smoke_raise_velo):
    #     smoke_raise_velo = lambda: smoke_raise_velo
    #
    # if not callable(fire_raise_velo):
    #     fire_raise_velo = lambda : fire_raise_velo

    smoke = Emitter(EmitterArgs(
        origin=lambda: origin() + Vector2((rnd.random() * 2 - 1) * 1, -75),
        velocity=smoke_raise_velo,
        destroyers=[partial(time_destroyer, lifespan_ms=smoke_lifespan_ms)],
        size=lambda x: utils.grow(x, lifespan_ms=smoke_lifespan_ms, max_size=smoke_end_size, start_size=smoke_end_size/4),
        particle_generation_args=smoke_generation_args,
        renderer=smoke_renderer
        ),
        id='torch_smoke'
    )


    flame = Emitter(EmitterArgs(
        origin=lambda: origin() + Vector2((rnd.random() * 2 - 1) * 5, 0),
        velocity=fire_raise_velo,
        destroyers=[partial(time_destroyer, lifespan_ms=fire_lifespan_ms)],
        size=lambda x: utils.decay(x, lifespan_ms=fire_lifespan_ms, max_size=fire_start_size),
        particle_generation_args=fire_generation_args,
        renderer=fire_renderer
       ),
        id='torch_flame'
    )

    torch = EmitterComposite(
        emitters={
            0: smoke,
            1: flame
        }

    )

    return torch


def twinkle_stars(star_gen_args: ParticleGenerationArgs,
                  width_bounds: Tuple[int, int],
                  height_bounds: Tuple[int, int],
                  star_size: int,
                  twinkle_time_ms: Union[int, Callable[[], int]],
                  renderer: ParticleRenderer
                  ):

    origin = lambda: Vector2.from_tuple((rnd.randint(*width_bounds), rnd.randint(*height_bounds)))

    return Emitter(EmitterArgs(
        origin=origin,
        velocity=Vector2(0, 0),
        destroyers=[partial(time_and_size_destroyer, lifespan_ms=twinkle_time_ms, size_compare_args=CompareArgs(le=0))],
        size=lambda x: utils.sin_wave(x + rnd.random(), utils.SinWaveArgs(amp=star_size,wavelength=twinkle_time_ms)),
        particle_generation_args=star_gen_args,
        renderer=renderer
        ),
        id='stars'
    )


def fountain(origin: Vector2,
             spray_intensity: float,
             spray_arc: Tuple[int, int],
             fountain_lifespan_ms: int,
             drop_accel: Vector2,
             droplet_size_start: int,
             droplet_size_end: int,
             droplet_renderer: ParticleRenderer,
             droplet_generation_args: ParticleGenerationArgs = None
             ) -> UpdateRenderableProtocol:

    if droplet_generation_args is None:
        droplet_generation_args = ParticleGenerationArgs(generation_interval_ms=50)

    spray_dir_lam = lambda: Vector2.point_on_circle(rnd.randint(spray_arc[0], spray_arc[1]), center=(0, 0)) * - spray_intensity

    return Emitter(EmitterArgs(
        origin=lambda: origin,
        velocity=spray_dir_lam,
        destroyers=[partial(time_destroyer, lifespan_ms=fountain_lifespan_ms)],
        size=lambda x: utils.decay(x, lifespan_ms=fountain_lifespan_ms, max_size=droplet_size_start, end_size=droplet_size_end),
        particle_generation_args=droplet_generation_args,
        renderer=droplet_renderer,
        accel=drop_accel
        ),
        id='fountain'
    )

def firework(origin: Union[Vector2, vec2_callable],
             launch_speed: Union[float, float_callable],
             spread: Union[float, float_callable],
             lifespan_ms: int,
             drop_accel: Vector2,
             size_start: int,
             size_end: int,
             arms: Union[int, int_callable],
             burst_interval_ms: Union[int, int_callable],
             explosion_renderer: ParticleRenderer,
             launch_renderer: ParticleRenderer,
             spread_arc: Tuple[int, int] = None,
             max_fireworks: int = None,
             launch_start_callback: phase_callback = None,
             launch_end_callback: phase_callback = None,
             explosion_start_callback: phase_callback = None,
             explosion_end_callback: phase_callback = None
             ) -> UpdateRenderableProtocol:

    if callable(spread):
        spread = spread()

    if callable(launch_speed):
        launch_speed = launch_speed()

    spread_arc_lam = lambda min, max: (rnd.randint(min, max), rnd.randint(min, max))
    spray_dir_lam = lambda min, max, speed: Vector2.point_on_circle(rnd.randint(*sorted(spread_arc_lam(min, max))),
                                                    center=(0, 0)) * - launch_speed


    launch_lam = lambda x: Emitter(EmitterArgs(
        origin=origin,
        velocity=lambda: spray_dir_lam(60, 120, launch_speed),
        accel=Vector2(0, 100),
        destroyers=[lambda part: part.velo_u_s.y > 0],
        size=3,
        renderer=launch_renderer,
        particle_generation_args=ParticleGenerationArgs(generation_interval_ms=75, max_particles=1)),
        id='firework_launch',
        trail_args=EmitterArgs(
            origin=Vector2(0, 0),
            velocity=Vector2(0, 0),
            destroyers=[partial(time_destroyer, lifespan_ms=300)],
            size=1,
            renderer=launch_renderer,
            particle_generation_args=ParticleGenerationArgs(generation_interval_ms=75)
        )
    )

    explosion_generation_args = ParticleGenerationArgs(
        generation_interval_ms=50,
        burst_args=BurstArgs(
            burst_interval_ms=burst_interval_ms,
            particles_per_burst=arms,
            max_bursts=1
        )
    )


    explosion_lam = lambda x: Emitter(EmitterArgs(
        origin=lambda: x.LastPos,
        velocity=lambda: spray_dir_lam(0, 360, spread),
        destroyers=[partial(time_destroyer, lifespan_ms=lifespan_ms)],
        size=lambda x: utils.decay(x, lifespan_ms=lifespan_ms, max_size=size_start, end_size=size_end),
        particle_generation_args=explosion_generation_args,
        renderer=explosion_renderer,
        accel=drop_accel
        ),
        trail_args=EmitterArgs(
            origin=Vector2(0, 0),
            velocity=Vector2(0, 0),
            destroyers=[partial(time_destroyer, lifespan_ms=300)],
            size=1,
            renderer=explosion_renderer,
            particle_generation_args=ParticleGenerationArgs(generation_interval_ms=75)
        ),
        id='firework_explosion'
    )

    return Emitter(
        EmitterArgs(
            origin=origin,
            velocity=Vector2(0, 0),
            particle_generation_args=ParticleGenerationArgs(
                phase_args=PhaseArgs(
                    phases=[
                        EmitterPhase(emitter_provider=launch_lam,
                                     phase_end_trigger=lambda me: len(me.Particles) == 0 and me.LifetimeMS > 100,
                                     phase_start_callback=launch_start_callback,
                                     phase_end_callback=launch_end_callback), #lambda x: x.LifetimeMS > rnd.randint(1000, 2500)
                        EmitterPhase(emitter_provider=explosion_lam,
                                     phase_end_trigger=lambda me: me.LifetimeMS > lifespan_ms,
                                     phase_start_callback=explosion_start_callback,
                                     phase_end_callback=explosion_end_callback)
                    ],
                    max_phasers=max_fireworks
                ),
                generation_interval_ms=burst_interval_ms
            ),
            destroyers=[partial(time_destroyer, lifespan_ms=5000)],
            renderer=explosion_renderer,
            size=1
        ),
        id='fireworks',
        context='fireworks'
    )


from coopgame.characterDevelopment.behaviorExtensions import hp as hp
from coopgame.characterDevelopment import character as char
from coopgame.characterDevelopment import stats as stat
from typing import Callable, Tuple
import logging
import random as rnd

logger = logging.getLogger(__name__)

STRENGTH = 'STRENGTH'
SPEED = 'SPEED'
MARTIAL_COMBAT = 'MARTIAL_COMBAT'
FORTUNE = 'FORTUNE'



class DualStartedStatEvent(stat.DeltaStatStatsPackageStateEvent):
    def __init__(self, opponent: char.Character):


        super().__init__(

        )



DualingEpochHandler=Callable[[char.Character, char.Character],
                                Tuple[char.Character, char.Character]]



def basic_attack(attacker: char.Character,
                 defender: char.Character) -> Tuple[char.Character, char.Character]:
    attacker_stats = attacker.stats.State.InternallyResolvedStats.stats
    defender_stats = defender.stats.State.InternallyResolvedStats.stats

    fortune_window = 0.5


    attack_score = attacker_stats[STRENGTH].val * rnd.uniform(1 - fortune_window * 0.5 * (1 + attacker_stats[FORTUNE].val / 100),
                                                              1 + fortune_window * 0.5 * (1 + attacker_stats[FORTUNE].val / 100))
    defender_score = defender_stats[STRENGTH].val * rnd.uniform(1 - fortune_window * 0.5 * (1 + defender_stats[FORTUNE].val / 100),
                                                              1 + fortune_window * 0.5 * (1 + defender_stats[FORTUNE].val / 100))

    damage = int(max(0.0, attack_score - 0.75 * defender_score))

    attacker_speed = attacker_stats[SPEED].val
    defender_speed = defender_stats[SPEED].val

    p_miss = max(0.0, (defender_speed - attacker_speed) / 100)

    if rnd.random() > p_miss:
        hp.damage(defender, damage)

    return attacker, defender

def basic_dualing_epoch_handler(char1: char.Character, char2: char.Character) -> Tuple[char.Character, char.Character]:
    basic_attack(char1, char2)
    if hp.alive(character=char2):
        basic_attack(char2, char1)

    return char1, char2


def dual(char1: char.Character,
           char2: char.Character,
           dual_epoch_handler: DualingEpochHandler) -> Tuple[char.Character, char.Character]:
    while True:
        char1, char2 = dual_epoch_handler(char1, char2)

        dead_chars = []
        for char in [char1, char2]:
            if hp.dead(character=char):
                logger.info(f"Character: {char.name} was killed!")
                dead_chars.append(char)

        if len(dead_chars) > 0:
            break

    return char1, char2









from coopgame.gameTemplate import GameTemplate
from cooptools.colors import Color
from coopgame.handleKeyPressedArgs import InputState, InputStateHandler, InputAction, InputEvent, InputEventType, basic_key_actions, CallbackPackage
from coopgame.pygame_k_constant_names import PyKeys, PyMouse
from coopgraph.gridgraph import GridGraph
from coopgame.grids_graphs.pygridhandler import PyGridHandler, GridSurfaceType
from coopgame.grids_graphs.pygraphhandler import PyGraphHandler, GraphSurfaceType, DEFAULT_DRAW_CONFIG
from cooptools.sectors.grids import RectGrid
from cooptools.dictPolicies import TogglePolicy
from cooptools.toggles import BooleanToggleable
from cooptools.coopEnum import CardinalPosition
from coopstructs.geometry.vectors.vectorN import Vector2

ROWS = 21
COLS = 21
TOGGLED_KEY = 'toggled'
GRID_TOGGLE_POLICY = TogglePolicy(key=TOGGLED_KEY, toggle_type=BooleanToggleable, default=False)

class GridGraphGame_Basic(GameTemplate):

    def __init__(self):
        super().__init__(input_state_handler=self.define_input_handler())

        self.grid_graph = GridGraph(
            grid=RectGrid(
                nRows=ROWS,
                nColumns=COLS,
            ),
            connect_diagonals=False,
            graph_node_enabled_toggle_key=TOGGLED_KEY
        )

        self.grid_handler = PyGridHandler(
            screen=self.screen,
            grid_provider=lambda: self.grid_graph.grid,
            get_game_area_rect_callback=lambda: self.game_area_rectangle,
            # get_highlights_callback=lambda: STATIC_HIGHLIGHTED_CELLS,
            # get_highlighted_rows=lambda: self.highlighted_rows(),
            # get_highlighted_cols=lambda: self.highlighted_cols(),
            get_toggled_key_colors=lambda: {GRID_TOGGLE_POLICY.key: Color.RED}
        )
        self.graph_handler = PyGraphHandler(
            screen=self.screen,
            graph=self.grid_graph.graph,
            draw_config=DEFAULT_DRAW_CONFIG.with_(
                coordinate_converter=lambda x: Vector2.from_tuple(self.grid_graph.grid.coord_from_grid_pos(
                    grid_pos=x.as_tuple(),
                    cardinal_pos=CardinalPosition.CENTER,
                    area_wh=self.game_area_rectangle.Dims)
                ))
        )

    def define_input_handler(self):
        return InputStateHandler(
            quit_callback_package=CallbackPackage(down=lambda x: self.quit()),
            debug_callback_package=CallbackPackage(down=lambda x: self.debug_mode_toggle.toggle()),
            fullscreen_callback_package=CallbackPackage(down=lambda x: self.toggle_fullscreen()),
            key_actions=[
                InputAction([InputEvent(PyKeys.K_r, event_type=InputEventType.DOWN)], callback=lambda x: self.redraw()),
                InputAction([InputEvent(PyKeys.K_1, event_type=InputEventType.DOWN)], callback=lambda x: self.grid_handler.toggle_surface([GridSurfaceType.GRID_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_2, event_type=InputEventType.DOWN)], callback=lambda x: self.grid_handler.toggle_surface([GridSurfaceType.HIGHLIGHTS_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_3, event_type=InputEventType.DOWN)], callback=lambda x: self.graph_handler.toggle_surface([GraphSurfaceType.EDGES_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_4, event_type=InputEventType.DOWN)], callback=lambda x: self.graph_handler.toggle_surface([GraphSurfaceType.NODES_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_5, event_type=InputEventType.DOWN)], callback=lambda x: self.graph_handler.toggle_surface([GraphSurfaceType.EDGE_LABELS_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_6, event_type=InputEventType.DOWN)], callback=lambda x: self.graph_handler.toggle_surface([GraphSurfaceType.NODE_LABELS_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_7, event_type=InputEventType.DOWN)], callback=lambda x: self.graph_handler.toggle_surface([GraphSurfaceType.OVERLAY_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_KP_PLUS, event_type=InputEventType.DOWN)], callback=lambda x: self.show_all()),
                InputAction([InputEvent(PyKeys.K_KP_MINUS, event_type=InputEventType.DOWN)], callback=lambda x: self.hide_all()),
                InputAction([InputEvent(PyMouse.LEFT, event_type=InputEventType.DOWN)], callback=lambda x: self.toggle_grid_at_mouse_pos()),

        ])


    def redraw(self):
        self.grid_handler.redraw()
        self.graph_handler.redraw()

    def hide_all(self):
        self.graph_handler.hide_all()
        self.grid_handler.hide_all()

    def show_all(self):
        self.graph_handler.show_all()
        self.grid_handler.show_all()


    def handle_left_click(self):
        grid_pos = self.grid_handler.mouse_grid_pos()

        if self.grid_handler.MouseHoverGridPos is not None:
            self.grid_graph.act_on_grid(
                row=self.grid_handler.MouseHoverGridPos[0],
                column=self.grid_handler.MouseHoverGridPos[1],
                policies=[GRID_TOGGLE_POLICY]
            )


    def toggle_grid_at_mouse_pos(self):
        mgpo = self.grid_handler.mouse_grid_pos()
        self.grid_graph.act_on_grid(row=int(mgpo.x), column=int(mgpo.y), policies=[GRID_TOGGLE_POLICY])
        self.grid_handler.invalidate()
        self.graph_handler.invalidate()

    def draw(self, frames: int, debug_mode: bool = False):
        self.grid_handler.render(self.screen)
        self.graph_handler.render(self.screen)


if __name__ == "__main__":
    g = GridGraphGame_Basic()
    g.main()


from typing import List

from coopgame.gameTemplate import GameTemplate, BuiltInSurfaceType
from coopgame.spriteHandling.spritePoseHandler import SpritePoseHandler
from coopgame.spriteHandling.spriteFolder import SpriteFolderHandler
from coopgame.spriteHandling.sprites import AnimatedPoseSprite
from coopgame.spriteHandling.spritesheet import SpriteSheet
from cooptools.toggles import IntegerRangeToggleable
import cooptools.geometry_utils.vector_utils as vec
import pygame
import coopgame.pygamehelpers as help
from coopgame.label_drawing import label_drawing_utils as ldu
from coopgame.label_drawing.pyLabel import TextAlignmentType
from cooptools.coopEnum import CardinalPosition
from cooptools.colors import Color
from coopgame.handleKeyPressedArgs import InputState, InputStateHandler, InputAction, InputEvent, InputEventType, basic_key_actions, CallbackPackage
from coopgame.pygame_k_constant_names import PyKeys, PyMouse

SPRITE_DIRS = [
    r'C:\Users\tburns\Downloads\ilyGuy',
    # r'C:\Users\tburns\Downloads\ninjaadventurenew\png',
    # r'C:\Users\tburns\Downloads\freeknight\png',
    # r'C:\Users\tburns\Downloads\zombiefiles\png\male',
    # r'C:\Users\tburns\Downloads\zombiefiles\png\female',
    # r'C:\Users\tburns\Downloads\craftpix-net-786503-free-spaceship-pixel-art-sprite-sheets\Fighter'
]




class SpriteTester(GameTemplate):

    def __init__(self):
        self.animation_cycle = IntegerRangeToggleable(min=100, max=1000, step_size=25)

        input_state_handler = InputStateHandler(
            quit_callback_package=CallbackPackage(down=lambda x: self.quit()),
            debug_callback_package=CallbackPackage(down=lambda x: self.debug_mode_toggle.toggle()),
            fullscreen_callback_package=CallbackPackage(down=lambda x: self.toggle_fullscreen()),
            key_actions=[
                InputAction([InputEvent(PyKeys.K_DOWN, event_type=InputEventType.DOWN)], callback=lambda x: self.animation_cycle.toggle(reverse=True)),
                InputAction([InputEvent(PyKeys.K_UP, event_type=InputEventType.DOWN)], callback=lambda x: self.animation_cycle.toggle()),
                InputAction([InputEvent(PyKeys.K_LEFT, event_type=InputEventType.DOWN)], callback=lambda x: self.Sprite.pose_handler.decrement_pose()),
                InputAction([InputEvent(PyKeys.K_RIGHT, event_type=InputEventType.DOWN)], callback=lambda x: self.Sprite.pose_handler.increment_pose())
        ])

        super().__init__(input_state_handler=input_state_handler)
        # self.sprite_handler = SpritePoseHandler.from_spritefolder(
        #     SpriteFolderHandler(SPRITE_DIRS[4]))
        self.sprite_handler = SpritePoseHandler.from_spritesheet(
            sprite_sheet=SpriteSheet(filename=r'C:\Users\Tj Burns\OneDrive - T-Hive\Desktop\ilyGuy.png',
                                     n_columns=2,
                                     n_rows=3,
                                     x_margin_right=32
                                     ),
            pose_def={
                'waving': (0, 0, 2),
                'idle': (1, 0, 2),
                'walkright': (2, 0, 3)
            }
        )

        self.anim_sprite = AnimatedPoseSprite(
            id='main',
            init_pos=(0, 0),
            pose_handler=self.sprite_handler,
            animation_cycle_ms=lambda: self.animation_cycle.value,
            width=int(self.screen_width / 3),
            height=int(self.screen_height)
        )

    def initialize_game(self):
        pass

    def draw(self, frames: int, debug_mode: bool = False):
        self.anim_sprite.blit(self.screen)
        current_pose_data = self.anim_sprite.pose_handler.get_current()
        ldu.draw_label(self.screen,
                       f"Current Pose: {current_pose_data['pose']}, "
                       f"anim_idx: {current_pose_data['animation_idx']}, "
                       f"anim cycle: {self.animation_cycle.value}",
                       args=ldu.DrawLabelArgs(
                           color=Color.CYAN,
                           font=ldu.DEFAULT_FONT,
                           alignment=TextAlignmentType.TOPLEFT,
                           anchor_alignment=CardinalPosition.TOP_LEFT
                       ),
                       pos=(50, 50))


    def model_updater(self, delta_time_ms: int):
        pass

    def sprite_updater(self, delta_time_ms: int):
        self.anim_sprite.animate(delta_time_ms)

    def on_resize(self):
        pass

    def _increase_animation_cycle(self):
        self.animation_cycle.toggle(loop=False)

    def _decrease_animation_cycle(self):
        self.animation_cycle.toggle(reverse=True, loop=False)

    def update_built_in_surfaces(self, surface_types: List[BuiltInSurfaceType]):
        pass

    @property
    def Sprite(self):
        return self.anim_sprite

if __name__ == "__main__":
    import logging

    logging.basicConfig(level=logging.DEBUG)

    game = SpriteTester()
    game.main()


import coopgame.characterDevelopment.stats as stat
from cooptools.coopEnum import CoopEnum, auto
from cooptools.common import verify_val, bound_value, verify
from coopgame.characterDevelopment.manifests import MutuallyExclusiveManifest

###################
# stats
###################
class BuiltInStatType(CoopEnum):
    HEIGHT_M = auto()  # General attribute
    WEIGHT_KG = auto()  # General attribute
    HP = auto()  # General attribute
    # MAX_HP = auto() # General attribute # TODO: How to handle this correctly? Its both part of meta AND a good stat to track. How will updating HP honor this stat?
    EXPERIENCE = auto()  # General attribute
    INTELLECT = auto()  # affects how the character is capable of figuring out problems using intuition
    EMOTION = auto()  # affects how the character is capable of figuring out problems using feeling
    CHARISMA = auto()  # affects how the character interacts with other characters and their willingness to accept the character
    PREMONITION = auto()  # affects if the character can "sense" things that are about to happen
    OBSERVATION = auto()  # affects if the character can notice details in the environment
    STRENGTH = auto()  # physical ability to perform heavy feats
    NIMBLENESS = auto()  # physical ability to perform complex feats
    SPEED = auto()  # physical ability to perform fast feats
    MAGIC = auto()  # "strength" of the magic order
    MANA = auto()  # amount of available pool to cast spells
    ATTRACTIVENESS = auto()  # affects how other characters perceive the character
    LEADERSHIP = auto()  # affects the ability of the character to maintain control over other characters
    MARTIAL_TACTICS = auto() # affects the ability of the character to perform combat related decisions
    MARTIAL_STRATEGY = auto()  # affects the ability of the character to perform war planning related decisions
    RATIONALITY = auto()  # affects the ability of the character to use 'INTELLECT' rather than 'EMOTION' when making decisions
    AMBITION = auto()  # affects the characters likelihood to strive for better position
    BOLDNESS = auto()  # affects the characters likelihood to take risks
    HEALTH = auto() # describes the characters overall well-being and likelihood to contract ailments
    FORTUNE = auto() # affects many other attributes when

STAT_DEFAULTS = {
    BuiltInStatType.HEIGHT_M: stat.StatState(stat_name=BuiltInStatType.HEIGHT_M.name, val=2, meta=stat.StatMeta(min=0.25, max=5)),
    BuiltInStatType.WEIGHT_KG: stat.StatState(stat_name=BuiltInStatType.WEIGHT_KG.name, val=75, meta=stat.StatMeta(min=25, max=250)),
    BuiltInStatType.SPEED: stat.StatState(stat_name=BuiltInStatType.SPEED.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.HP: stat.StatState(stat_name=BuiltInStatType.HP.name, val=1000, meta=stat.StatMeta(min=0, max=1000)),
    BuiltInStatType.INTELLECT: stat.StatState(stat_name=BuiltInStatType.INTELLECT.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.CHARISMA: stat.StatState(stat_name=BuiltInStatType.CHARISMA.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.PREMONITION: stat.StatState(stat_name=BuiltInStatType.PREMONITION.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.OBSERVATION: stat.StatState(stat_name=BuiltInStatType.OBSERVATION.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.MAGIC: stat.StatState(stat_name=BuiltInStatType.MAGIC.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.MANA: stat.StatState(stat_name=BuiltInStatType.MANA.name, val=100, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.EXPERIENCE: stat.StatState(stat_name=BuiltInStatType.EXPERIENCE.name, val=0, meta=stat.StatMeta(min=0)),
    BuiltInStatType.STRENGTH: stat.StatState(stat_name=BuiltInStatType.STRENGTH.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.ATTRACTIVENESS: stat.StatState(stat_name=BuiltInStatType.ATTRACTIVENESS.name, val=10, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.EMOTION: stat.StatState(stat_name=BuiltInStatType.EMOTION.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.NIMBLENESS: stat.StatState(stat_name=BuiltInStatType.NIMBLENESS.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.LEADERSHIP: stat.StatState(stat_name=BuiltInStatType.LEADERSHIP.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.MARTIAL_TACTICS: stat.StatState(stat_name=BuiltInStatType.MARTIAL_TACTICS.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.MARTIAL_STRATEGY: stat.StatState(stat_name=BuiltInStatType.MARTIAL_STRATEGY.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.RATIONALITY: stat.StatState(stat_name=BuiltInStatType.RATIONALITY.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.AMBITION: stat.StatState(stat_name=BuiltInStatType.AMBITION.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.BOLDNESS: stat.StatState(stat_name=BuiltInStatType.BOLDNESS.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.HEALTH: stat.StatState(stat_name=BuiltInStatType.HEALTH.name, val=15, meta=stat.StatMeta(min=0, max=100)),
    BuiltInStatType.FORTUNE: stat.StatState(stat_name=BuiltInStatType.FORTUNE.name, val=0, meta=stat.StatMeta(min=-25, max=25))
}

def _verify_stat_defaults():
    verify(lambda: all(x in STAT_DEFAULTS.keys() for x in BuiltInStatType), msg=f"Missing trait(s) definitions: \n {[x for x in BuiltInStatType if x not in STAT_DEFAULTS.keys()]}")
_verify_stat_defaults()


##################
# skills
##################
class Skill(CoopEnum):
    WOODCUTTER = auto()

skills_manifest = {
    Skill.WOODCUTTER: stat.StatEffectPackage(id=Skill.WOODCUTTER.name, description="Cut wood",
                                             stat_effects=frozenset([stat.StatEffect(BuiltInStatType.STRENGTH.name, delta_effect=5)]),
                                           category='SKILL')
}

def _verify_skill_manifest():
    verify(lambda: all(x in skills_manifest.keys() for x in Skill), msg=f"Missing trait(s) definitions: \n {[x for x in Skill if x not in skills_manifest.keys()]}")
_verify_skill_manifest()


###################
# Personality traits
###################
class PersonalityTrait(CoopEnum):
    CAUTIOUS = auto()
    CARELESS = auto()
    BRAVE = auto()

"""
Intended Trait effect relationships:
    Courage: Brave / Cowardly
    Sexuality: Chaste / Lustful
    Disturb: Peaceful / Excitable
    Ambition: Content / Ambitious
    WorkEthic: Diligent / Lazy
    Acceptance: Forgiving / Vengeful
    Giving: Generous / Greedy
    Charisma: Introverted / Extroverted
    Honesty: Truthful / Deceitful
    Humility: Humble / Arrogant
    Justice: Just / Arbitrary
    Patience: Patient / Impatient
    Paranoia: Trusting / Paranoid
    Fortune: Lucky / UnLucky
"""


personality_trait_manifest =  {
    PersonalityTrait.CAUTIOUS: stat.StatEffectPackage(id=PersonalityTrait.CAUTIOUS.name, description="affects observation premonition speed",
                                                      stat_effects=frozenset([stat.StatEffect(BuiltInStatType.PREMONITION.name, delta_effect=10),
                                                             stat.StatEffect(BuiltInStatType.OBSERVATION.name, delta_effect=10),
                                                             stat.StatEffect(BuiltInStatType.SPEED.name, delta_effect=-10)]),
                                           category='PERSONALITY_TRAIT'),
    PersonalityTrait.CARELESS: stat.StatEffectPackage(id=PersonalityTrait.CARELESS.name, description="affects observation premonition speed",
                                                      stat_effects=frozenset([stat.StatEffect(BuiltInStatType.PREMONITION.name, delta_effect=-10),
                                                             stat.StatEffect(BuiltInStatType.OBSERVATION.name, delta_effect=-10),
                                                             stat.StatEffect(BuiltInStatType.SPEED.name, delta_effect=10)]),
                                           category='PERSONALITY_TRAIT'),
    PersonalityTrait.BRAVE: stat.StatEffectPackage(id=PersonalityTrait.BRAVE.name, description="affects observation premonition speed",
                                                   stat_effects=frozenset([stat.StatEffect(BuiltInStatType.ATTRACTIVENESS.name, delta_effect=5),
                                                        stat.StatEffect(BuiltInStatType.CHARISMA.name, delta_effect=5),
                                                        stat.StatEffect(BuiltInStatType.AMBITION.name, delta_effect=5),
                                                        stat.StatEffect(BuiltInStatType.BOLDNESS.name, delta_effect=15),
                                                        stat.StatEffect(BuiltInStatType.LEADERSHIP.name, delta_effect=10)]),
                                           category='PERSONALITY_TRAIT')
}

def _verify_personality_trait_manifest():
    verify(lambda: all(x in personality_trait_manifest.keys() for x in PersonalityTrait), msg=f"Missing trait(s) definitions: \n {[x for x in PersonalityTrait if x not in personality_trait_manifest.keys()]}")
_verify_personality_trait_manifest()



###################
# Physical Traits
###################
class PhysicalTraitType(CoopEnum):
    BEAUTY = auto()


class PhysicalTrait(CoopEnum):
    UGLY = auto()
    WEAK = auto()

"""
Intended Trait effect relationships:
    Beauty: Ugly / Plain / Attractive / Beautiful
    Strength: Puny / Weak / Strong / Mighty
    Height: Dwarf / Short / Tall / Giant
    Weight: Skinny / Thin / Stout / Obese


"""

physical_trait_manifest = {
    PhysicalTrait.UGLY: stat.StatEffectPackage(id=PhysicalTrait.UGLY.name, description="affects others perception of their beauty",
                                                  stat_effects=frozenset([stat.StatEffect(BuiltInStatType.ATTRACTIVENESS.name, delta_effect=-15)]),
                                           category='PHYSICAL_TRAIT'),
    PhysicalTrait.WEAK: stat.StatEffectPackage(id=PhysicalTrait.WEAK.name, description="affects strength",
                                                  stat_effects=frozenset([stat.StatEffect(BuiltInStatType.STRENGTH.name, delta_effect=-15)]),
                                           category='PHYSICAL_TRAIT'),
}


def _verify_personality_trait_manifest():
    verify(lambda: all(x in personality_trait_manifest.keys() for x in PersonalityTrait),
           msg=f"Missing trait(s) definitions: \n {[x for x in PersonalityTrait if x not in personality_trait_manifest.keys()]}")


_verify_personality_trait_manifest()


###################
# status effects
###################
class StatusEffect(CoopEnum):
    FATIGUED = auto()
    ENERGIZED = auto()
    PANICKED = auto()
    CALM = auto()

"""
Intended Status Efects:

"""
panicked = stat.StatEffectPackage(id=StatusEffect.PANICKED.name, description="affects strength and speed and charisma and observation",
                                                   stat_effects=frozenset([stat.StatEffect(BuiltInStatType.STRENGTH.name, delta_effect=10),
                                                                           stat.StatEffect(BuiltInStatType.SPEED.name, delta_effect=10),
                                                                           stat.StatEffect(BuiltInStatType.CHARISMA.name, delta_effect=-10),
                                                                           stat.StatEffect(BuiltInStatType.OBSERVATION.name, delta_effect=-20),
                                                                           stat.StatEffect(BuiltInStatType.BOLDNESS.name, delta_effect=-20),
                                                                           stat.StatEffect(BuiltInStatType.AMBITION.name, delta_effect=-20),
                                                                           stat.StatEffect(BuiltInStatType.ATTRACTIVENESS.name, delta_effect=-10),
                                                                           stat.StatEffect(BuiltInStatType.RATIONALITY.name, delta_effect=-15),
                                                                           stat.StatEffect(BuiltInStatType.MARTIAL_STRATEGY.name, delta_effect=-15),
                                                                           stat.StatEffect(BuiltInStatType.MARTIAL_TACTICS.name, delta_effect=-15),
                                                                           stat.StatEffect(BuiltInStatType.LEADERSHIP.name, delta_effect=-10),
                                                                           stat.StatEffect(BuiltInStatType.NIMBLENESS.name, delta_effect=-10),
                                                                           stat.StatEffect(BuiltInStatType.EMOTION.name, delta_effect=10),
                                                                           stat.StatEffect(BuiltInStatType.INTELLECT.name, delta_effect=-10)]),
                                           category='STATUS_EFFECT')
fatigued = stat.StatEffectPackage(id=StatusEffect.FATIGUED.name, description="affects strength and speed and charisma and observation",
                                                  stat_effects=frozenset([stat.StatEffect(BuiltInStatType.STRENGTH.name, delta_effect=-10),
                                                        stat.StatEffect(BuiltInStatType.SPEED.name, delta_effect=-10),
                                                        stat.StatEffect(BuiltInStatType.CHARISMA.name, delta_effect=-5),
                                                        stat.StatEffect(BuiltInStatType.OBSERVATION.name, delta_effect=-3)]),
                                           category='STATUS_EFFECT')
status_effect_manifest = {
    StatusEffect.FATIGUED: fatigued,
    StatusEffect.ENERGIZED: fatigued.with_opposite_stat_effects(id=StatusEffect.ENERGIZED.name, description="Energized"),
    StatusEffect.PANICKED: panicked,
    StatusEffect.CALM: panicked.with_opposite_stat_effects(id=StatusEffect.CALM.name, description="Calm")
}
def _verify_status_effect_manifest():
    verify(lambda: all(x in status_effect_manifest.keys() for x in StatusEffect), msg=f"Missing trait(s) definitions: \n {[x for x in StatusEffect if x not in status_effect_manifest.keys()]}")
_verify_status_effect_manifest()




########################
# Global Events
########################
class GlobalEffect(CoopEnum):
    FAMINE = auto()

global_effect_manifest = {
    GlobalEffect.FAMINE: stat.StatEffectPackage(id=GlobalEffect.FAMINE.name, description="Worldwide food shortage. Affects strength, speed, emotion and rationality",
                                                  stat_effects=frozenset([stat.StatEffect(BuiltInStatType.STRENGTH.name, delta_effect=-10),
                                                        stat.StatEffect(BuiltInStatType.SPEED.name, delta_effect=-10),
                                                        stat.StatEffect(BuiltInStatType.EMOTION.name, delta_effect=10),
                                                        stat.StatEffect(BuiltInStatType.RATIONALITY.name, delta_effect=-15)]),
                                           category='GLOBAL')
}


def _verify_global_effect_manifest():
    verify(lambda: all(x in global_effect_manifest.keys() for x in GlobalEffect), msg=f"Missing gloableffect(s) definitions: \n {[x for x in GlobalEffect if x not in global_effect_manifest.keys()]}")
_verify_global_effect_manifest()


########################
# Races
########################
class RaceType(CoopEnum):
    HUMAN = auto()
    ORC = auto()
    ELF = auto()
    DARK_ELF = auto()
    DWARF = auto()

races_manifest = {
    RaceType.HUMAN: stat.StatEffectPackage(id=RaceType.HUMAN.name, description="",
                                           stat_effects=frozenset([
                                                        stat.StatEffect(BuiltInStatType.EMOTION.name, delta_effect=10),
                                                        stat.StatEffect(BuiltInStatType.RATIONALITY.name, delta_effect=-5)
                                                  ]),
                                           category='RACE'),
    RaceType.ORC: stat.StatEffectPackage(id=RaceType.ORC.name, description="",
                                         stat_effects=frozenset([
                                                   stat.StatEffect(BuiltInStatType.STRENGTH.name, delta_effect=10),
                                                   stat.StatEffect(BuiltInStatType.SPEED.name, delta_effect=-10),
                                                   stat.StatEffect(BuiltInStatType.EMOTION.name, delta_effect=10),
                                                   stat.StatEffect(BuiltInStatType.RATIONALITY.name, delta_effect=-10),
                                                   stat.StatEffect(BuiltInStatType.MAGIC.name, delta_effect=-5),
                                                   stat.StatEffect(BuiltInStatType.WEIGHT_KG.name, delta_effect=20),
                                                   stat.StatEffect(BuiltInStatType.HEIGHT_M.name, delta_effect=.5),
                                               ]),
                                           category='RACE'),
    RaceType.ELF: stat.StatEffectPackage(id=RaceType.ELF.name, description="",
                                         stat_effects=frozenset([
                                                 stat.StatEffect(BuiltInStatType.STRENGTH.name, delta_effect=-10),
                                                 stat.StatEffect(BuiltInStatType.SPEED.name, delta_effect=15),
                                                 stat.StatEffect(BuiltInStatType.RATIONALITY.name, delta_effect=10),
                                                 stat.StatEffect(BuiltInStatType.MAGIC.name, delta_effect=5),
                                                 stat.StatEffect(BuiltInStatType.WEIGHT_KG.name, delta_effect=-20),
                                                 stat.StatEffect(BuiltInStatType.HEIGHT_M.name, delta_effect=.25),
                                             ]),
                                           category='RACE'),
    RaceType.DARK_ELF: stat.StatEffectPackage(id=RaceType.DARK_ELF.name, description="",
                                              stat_effects=frozenset([
                                                 stat.StatEffect(BuiltInStatType.STRENGTH.name, delta_effect=5),
                                                 stat.StatEffect(BuiltInStatType.SPEED.name, delta_effect=10),
                                                 stat.StatEffect(BuiltInStatType.MAGIC.name, delta_effect=-5),
                                                 stat.StatEffect(BuiltInStatType.WEIGHT_KG.name, delta_effect=-20),
                                                 stat.StatEffect(BuiltInStatType.HEIGHT_M.name, delta_effect=.25),
                                             ]),
                                           category='RACE'),
    RaceType.DWARF: stat.StatEffectPackage(id=RaceType.DWARF.name, description="",
                                           stat_effects=frozenset([
                                                      stat.StatEffect(BuiltInStatType.STRENGTH.name, delta_effect=15),
                                                      stat.StatEffect(BuiltInStatType.SPEED.name, delta_effect=-10),
                                                      stat.StatEffect(BuiltInStatType.MAGIC.name, delta_effect=-5),
                                                      stat.StatEffect(BuiltInStatType.WEIGHT_KG.name, delta_effect=15),
                                                      stat.StatEffect(BuiltInStatType.HEIGHT_M.name, delta_effect=-.5),
                                                  ]),
                                           category='RACE'),
}


def _verify_races_manifest():
    verify(lambda: all(x in races_manifest.keys() for x in RaceType),
           msg=f"Missing race(s) definitions: \n {[x for x in RaceType if x not in races_manifest.keys()]}")
_verify_races_manifest()



if __name__ == "__main__":
    pass






import coopgame.characterDevelopment.stats as stat
from coopgame.characterDevelopment import character as char

EXPERIENCE = 'EXPERIENCE'

def gain_experience(character: char.Character,
                    xp_amount: float,
                    experience_stat_name: str = EXPERIENCE):
    character.stats.register_events(events=[stat.DeltaStatStatsPackageStateEvent(stat_name=experience_stat_name, delta=xp_amount)])
    return character


import logging
import uuid
from typing import Optional
from cooptools.geometry_utils import vector_utils as vec
from dataclasses import dataclass, field
from cooptools.physics.kinematic import Transform
from cooptools.physics import waypointFollower as wf
from examples.voroni_world.constants import *
from examples.voroni_world.behaviorTypes import BehaviorType

logger = logging.getLogger(__name__)

@dataclass(kw_only=True)
class Villager:
    name: int
    waypoint_follower: wf.WaypointFollower = field(init=False)
    id: uuid.UUID = field(init=False, default_factory=uuid.uuid4)
    __behavior: Optional[BehaviorType] = None

    def __post_init__(self):
        self.waypoint_follower = wf.WaypointFollower(
            name=str(self.name),
            initial_transform=Transform(translation=(0, 0)),
            max_acceleration=VILLAGER_MAX_ACCEL,
            velocity_magnitude_bounds=(0, VILLAGER_MAX_VELO)
        )

    def set_pos(self, pos: vec.FloatVec):
        self.waypoint_follower.update(delta_time_ms=0, set_pos=pos)
        return self

    def set_behavior(self, behavior: BehaviorType):
        logger.info(f"Setting villager \'{self.name}\'s behavior to \'{behavior}\'")
        self.__behavior = behavior

    @property
    def Position(self) -> vec.FloatVec:
        return self.waypoint_follower.Position

    @property
    def Behavior(self) -> BehaviorType:
        return self.__behavior



import random as rnd
from coopgame.gameTemplate import GameTemplate, tt
from cooptools.decor import timer
import logging
from examples.voroni_world.models import *
from typing import List, Dict, Set, Tuple, Self, Iterator

logger = logging.getLogger(__name__)

@timer(logger=logger, log_level=0, time_tracking_class=tt)
def __wave_collapse_terrain(vn_map: Dict[int, Set[int]]) -> Dict[int, TerrainType]:
    ret = {r_id: None for r_id, neighbors in vn_map.items()}
    options = {r_id: [x for x in TerrainType] for r_id in vn_map.keys()}

    region_to_set = None
    while len({k: v for k, v in ret.items() if v is None}) > 0:
        if region_to_set is None:
            # any have only one option
            region_to_set = next((r_id for r_id, opts in options.items()
                           if opts is not None and len(opts) == 1), None)

            # choose random if none only have one option
            if region_to_set is None:
                region_to_set = rnd.choice([r_id for r_id, terr in ret.items() if terr is None])

        terrain_to_set = rnd.randint(0, len(options[region_to_set]) - 1)

        # set the choice
        ret[region_to_set] = options[region_to_set][terrain_to_set]

        # update options
        for neighbor in vn_map[region_to_set]:
            if options[neighbor] is None:
                continue

            # which options are no longer valid?
            incompatibles = [x for x in options[neighbor]
                             if x not in TERRAIN_COMPATIBLITY_MAP[ret[region_to_set]]]

            # remove them
            [options[neighbor].remove(x) for x in incompatibles]

            if len(options[neighbor]) == 0:
                blocks = {k: v for k, v in ret.items() if k in vn_map[neighbor]}
                msg = f"The options have been reduced to zero for region {neighbor}." \
                      + f"\n: Neighbors: {blocks}"
                # raise ValueError(msg)
                logger.warning(msg)
                ret[neighbor] = TerrainType.UNDEFINED
                options[neighbor] = None


        options[region_to_set] = None

        # find the neighbor with the min options remaining
        neighbor_len_options = [len(opts) for rid, opts in options.items()
                           if rid in vn_map[region_to_set] and opts is not None]

        min_options = None
        if len(neighbor_len_options) > 0:
            min_options = min(neighbor_len_options)

        if min_options is not None:
            region_to_set = next(x for x in vn_map[region_to_set]
                                 if options[x] is not None and len(options[x]) == min_options)
        else:
            region_to_set = None

    return ret

@timer(logger=logger, log_level=0, time_tracking_class=tt)
def __count_conflicts(r, vn_map, t_map, t):
    conflicts = 0
    for n in vn_map[r]:
        if t not in TERRAIN_COMPATIBLITY_MAP[t_map[n]]:
            conflicts += 1
    return conflicts

@timer(logger=logger, log_level=0, time_tracking_class=tt)
def __min_conflicts_terrain(vn_map: Dict[int, Set[int]], t_map: Dict[int, TerrainType] = None):
    regions = vn_map.keys()
    rnd.shuffle(list(regions))

    if t_map is None:
        t_map = {r: rnd.choice(list(TerrainType)) for r in vn_map.keys()}

    totConflicts = None
    iteration = 0
    while totConflicts is None or totConflicts > 0:
        totConflicts = 0
        iteration += 1
        replacements = []

        for r in regions:
            conflicts = __count_conflicts(r, vn_map, t_map, t_map[r])
            if conflicts > 0:
                totConflicts += 1
                replacements.clear()
                for t in TerrainType:
                    con = __count_conflicts(r, vn_map, t_map, t)
                    if (con <= conflicts) | (rnd.randint(0, 100) < 5):
                        replacements.append(t)

                t_map[r] = rnd.choice(replacements)
        logger.debug(f"Iteration: {iteration}, Total Conflicts: {totConflicts}")
    return t_map

@timer(logger=logger, log_level=0, time_tracking_class=tt)
def __post_process(vn_map: Dict[int, Set[int]], t_map: Dict[int, TerrainType]):
    regions = vn_map.keys()
    rnd.shuffle(list(regions))

    for r in regions:
        # if terrain is diff from all surrounding, change to a surrounding
        if not any(t_map[r]==t_map[x] for x in vn_map[r]):
            t_map[r] = rnd.choice([t_map[x] for x in vn_map[r]])
    return t_map

def wave_collapse_terrains(vn_map):
    terrains = __wave_collapse_terrain(vn_map)
    terrains = __min_conflicts_terrain(vn_map, terrains)
    terrains = __post_process(vn_map, terrains)

    return terrains


import pygame
from dataclasses import dataclass, asdict
from cooptools.colors import Color
from coopgame.label_drawing.pyLabel import TextAlignmentType
from cooptools.anchor import Anchor2D, CardinalPosition
import logging
from typing import Tuple, Dict
from coopstructs.geometry import Rectangle
import coopgame.pygamehelpers as help
from coopgame.pointdrawing import point_draw_utils as putils
from coopgame.surfaceManager import SurfaceGroup

logger = logging.getLogger(__name__)

DEFAULT_FONT = pygame.font.Font(None, 20)

@dataclass(frozen=True)
class DrawLabelArgs:
    color: Color = None
    font: pygame.font.Font = None
    alignment: TextAlignmentType = None
    anchor_alignment: CardinalPosition = None
    alpha: float = None
    offset: Tuple[int, int] = None

    def with_(self, **kwargs):
        kw = asdict(self)
        kw.update(kwargs)
        return DrawLabelArgs(**kw)


@dataclass(frozen=True)
class LabelArgs:
    text: str
    draw_args: DrawLabelArgs
    pos: Tuple[float, float] = None
    offset_rect: Rectangle = None
    anchor_point_draw_args: putils.DrawPointArgs = None


def get_font_size(
        font: pygame.font.Font,
        text: str
):
    return font.size(text)

def draw_label(hud: pygame.Surface,
               args: LabelArgs,
               deb: bool = False
               ):
    font = args.draw_args.font
    if font is None:
        font = DEFAULT_FONT

    if args.pos is None and args.offset_rect is None:
        raise ValueError(f"At least one of pos and rect cannot be none")

    offset_rect = None
    if args.offset_rect is None:
        pos = [int(x) for x in args.pos[:2]]
        if args.draw_args.offset:
            pos = pos[0] + args.draw_args.offset[0], pos[1] + args.draw_args.offset[1]

        anchor_alignment = args.draw_args.anchor_alignment if args.draw_args.anchor_alignment is not None else CardinalPosition.BOTTOM_LEFT

        offset_rect = Rectangle(
            pt=pos,
            dims=get_font_size(font, args.text),
            anchor_cardinality=anchor_alignment
        )


    else:
        offset_rect = args.offset_rect

    color=args.draw_args.color
    if color is None:
        color = Color.BLUE

    alignment = args.draw_args.alignment
    if alignment is None:
        alignment = TextAlignmentType.TOPLEFT

    rendered_txt = font.render(args.text, True, color.value)

    alpha = args.draw_args.alpha
    if alpha:
        rendered_txt.set_alpha(alpha * 100)

    if deb:
        help.draw_box(
            surface=hud,
            rect=offset_rect,
            outline_color=Color.PINK,
            anchor_draw_args=putils.DrawPointArgs(
                color=Color.HOT_PINK,
                radius=4
            ),
            corner_draw_args=putils.DrawPointArgs(
                color=Color.LIGHT_BLUE,
                radius=1
            )
        )
    elif args.anchor_point_draw_args is not None:
        putils.draw_points(
            points={args.pos: args.anchor_point_draw_args},
            surface=hud
        )

    try:
        rect = rendered_txt.get_rect()
        align_coords = alignment_coords_for_type_rect(alignment, offset_rect)
        setattr(rect, alignment.value, align_coords)
        hud.blit(rendered_txt, rect)
        return rendered_txt
    except Exception as e:
        logger.error(f"{e}")

def alignment_coords_for_type_rect(alignment: TextAlignmentType, rect: Rectangle) -> Tuple[float, float]:
    """
    Note the orientation shift from top to bottom bc of the pygame inversion
    """
    switch = {
        TextAlignmentType.TOPRIGHT:      lambda: rect.planar.TopRight,
        TextAlignmentType.TOPLEFT:       lambda: rect.planar.TopLeft,
        TextAlignmentType.TOPCENTER:     lambda: rect.planar.TopCenter,
        TextAlignmentType.BOTTOMLEFT:    lambda: rect.planar.BottomLeft,
        TextAlignmentType.BOTTOMRIGHT:   lambda: rect.planar.BottomRight,
        TextAlignmentType.RIGHTCENTER:   lambda: rect.planar.RightCenter,
        TextAlignmentType.BOTTOMCENTER:  lambda: rect.planar.BottomCenter,
        TextAlignmentType.LEFTCENTER:    lambda: rect.planar.LeftCenter,
        TextAlignmentType.CENTER:        lambda: rect.Center
    }

    return switch.get(alignment)()

def draw_dict(dict_to_draw: Dict,
              draw_args: DrawLabelArgs,
              surface: pygame.Surface,
              total_game_time_sec: float,
              title: str = None,
              inter_line_buffer: int = 3,
              deb: bool=False):

    #TODO: This is too opinionated toward drawing game time. Should be more generic

    tracked_time = [(key, val) for key, val in dict_to_draw.items()]
    tracked_time.sort(key=lambda x: x[1], reverse=True)

    txt_lmbda = lambda key, val: f"{key}: {round(val, 2)} sec ({round(val / total_game_time_sec * 100, 1)}%)"

    rendered_txt = draw_label(
        hud=surface,
        args=LabelArgs(
            text=f"{title} {round(total_game_time_sec, 1)}s",
            draw_args=draw_args,
            pos=(0, 0),
        ),
        deb=deb
    )

    y_off = rendered_txt.get_rect().height + inter_line_buffer
    for key, val in tracked_time:

        rendered_txt = draw_label(
            hud=surface,
            args=LabelArgs(
                text=txt_lmbda(key, val),
                draw_args=draw_args,
                pos=(0, y_off),
            ),
            deb=deb
        )

        y_off += rendered_txt.get_rect().height + inter_line_buffer

    return y_off


def label_surface_group(
        labels: Dict[str, str],
        positions: Dict[str, Tuple[float, float]] = None,
        draw_label_args: DrawLabelArgs | Dict[str, DrawLabelArgs] = None
) -> SurfaceGroup:
    if draw_label_args is None:
        return None

    surf_lookups = {}
    surface_pos = {}
    for id, label in labels.items():
        if type(draw_label_args) == DrawLabelArgs:
            args = draw_label_args
        else:
            args = draw_label_args[id]

        if args is None:
            continue

        font = args.font
        if font is None:
            font = DEFAULT_FONT

        surf = help.init_surface(get_font_size(font, text=label))

        draw_label(surf,
                   args=LabelArgs(
                       text=label,
                       draw_args=args,
                       pos=(0, 0)
                   ))
        surf_lookups[id] = surf

        surface_pos[id] = positions.get(id, (0, 0)) if positions is not None else (0, 0)

    return SurfaceGroup(surfaces_lookup=surf_lookups,
                        surface_pos=surface_pos
                        )


from typing import List, Tuple
import random as rnd
import pygame.draw
import pygame
from coopstructs.geometry.vectors.vectorN import Vector2
from coopgame.gameTemplate import GameTemplate, BuiltInSurfaceType
from coopgame.chunkRendering import ChunkRenderer, chunk_of_tile
from coopgame.tilemap import Tilemap, FileTileset
from cooptools.colors import Color
import coopgame.pygamehelpers as help
from coopstructs.geometry import Rectangle
from cooptools.toggles import BooleanToggleable
from coopgame.spriteHandling.sprites import AnimatedSprite, SpriteOriginPosition
from coopgame.spriteHandling.spritesheet import SpriteSheet
from coopgame.handleKeyPressedArgs import InputState, InputStateHandler, InputAction, InputEvent, InputEventType, basic_key_actions, CallbackPackage
from coopgame.pygame_k_constant_names import PyKeys, PyMouse
from coopgame.scrollHandler import ScrollTileHandler
from cooptools.sectors import rect_utils as sec_util

file = '../coopgame/assets/tileSets/tmw_desert_spacing.png'

DELTA_AMT_PER_100MS = 20
WIDTH_DIM_P = (0.25, 0.65)
HEIGHT_DIM_P = (0.15, 0.85)
BG_TILE_STATE_HEADER = 'tileType'
TILE_HAS_PRIZE_HEADER = "hasPrize"

def dummy_tile_type_resolver(target: Tuple[int, int]) -> int:
    # thresh = 15
    # if target[1] > thresh:
    #     return 7 # underground
    # if target[1] == thresh:
    #     return 6 # surface
    # if target[1] == thresh - 1:
    #     if rnd.randint(1, 5) == 1:
    #         return rnd.choice([34, 39, 40, 41, 46, 47]) # foliage
    if target == (0, 0):
        return 35  # flag
    if rnd.randint(1, 20) == 1:
        return rnd.choice([34, 39, 40, 41, 46, 47])  # foliage
    return 33  # desert

def dummy_has_goal_resolver(target: Tuple[int, int]) -> int:
    if rnd.randint(1, 100) == 1:
        return 1  # has goal
    return 0 # no goal

class ChunkGame(GameTemplate):

    def __init__(self):
        input_state_handler = InputStateHandler(
            quit_callback_package=CallbackPackage(down=lambda x: self.quit()),
            debug_callback_package=CallbackPackage(down=lambda x: self.debug_mode_toggle.visibility_toggle()),
            fullscreen_callback_package=CallbackPackage(down=lambda x: self.toggle_fullscreen()),
            key_actions=[
                InputAction([InputEvent(PyKeys.K_LEFT, event_type=InputEventType.HELD)], callback=lambda key_pressed_args: self.head_left(key_pressed_args.DeltaMs)),
                InputAction([InputEvent(PyKeys.K_a, event_type=InputEventType.HELD)], callback=lambda key_pressed_args: self.head_left(key_pressed_args.DeltaMs)),
                InputAction([InputEvent(PyKeys.K_RIGHT, event_type=InputEventType.HELD)], callback=lambda key_pressed_args: self.head_right(key_pressed_args.DeltaMs)),
                InputAction([InputEvent(PyKeys.K_d, event_type=InputEventType.HELD)], callback=lambda key_pressed_args: self.head_right(key_pressed_args.DeltaMs)),
                InputAction([InputEvent(PyKeys.K_UP, event_type=InputEventType.HELD)], callback=lambda key_pressed_args: self.head_up(key_pressed_args.DeltaMs)),
                InputAction([InputEvent(PyKeys.K_w, event_type=InputEventType.HELD)], callback=lambda key_pressed_args: self.head_up(key_pressed_args.DeltaMs)),
                InputAction([InputEvent(PyKeys.K_DOWN, event_type=InputEventType.HELD)], callback=lambda key_pressed_args: self.head_down(key_pressed_args.DeltaMs)),
                InputAction([InputEvent(PyKeys.K_s, event_type=InputEventType.HELD)], callback=lambda x: self.head_down(x.DeltaMs)),
                InputAction([InputEvent(PyKeys.K_o, event_type=InputEventType.DOWN)], callback=lambda x: self.minimap_show_chunk_outline_toggle.toggle()),
                InputAction([InputEvent(PyKeys.K_p, event_type=InputEventType.DOWN)], callback=lambda x: self.minimap_show_player_chunk_toggle.toggle()),
                InputAction([InputEvent(PyKeys.K_z, event_type=InputEventType.DOWN)], callback=lambda x: input()),
                InputAction([InputEvent(PyKeys.K_t, event_type=InputEventType.DOWN)], callback=lambda x: self.minimap_show_player_tile_toggle.toggle()),
        ])

        super().__init__(screen_height=640, screen_width=960, input_state_handler=input_state_handler)
        self.tileset = FileTileset(file)
        self.tilemap = None

        pygame.display.set_caption("Pygame Tiled Demo")
        self.image = None

        self.chunk_renderer = ChunkRenderer(chunk_size=(16, 16), tile_type_resolvers={BG_TILE_STATE_HEADER: dummy_tile_type_resolver,
                                                                                      TILE_HAS_PRIZE_HEADER: dummy_has_goal_resolver})
        self.scroll = (0, 0)

        self.minimap_surface = pygame.Surface(size=(self.screen_width / 4,self.screen_height / 3))
        self.minimap_show_player_chunk_toggle = BooleanToggleable(default=False)
        self.minimap_show_player_tile_toggle = BooleanToggleable(default=False)
        self.minimap_show_chunk_outline_toggle = BooleanToggleable(default=False)

        character_file = '../coopgame/assets/spriteSheets/DA_male2.png'
        self.character_spritesheet = SpriteSheet(character_file, n_rows=21, n_columns=13, x_margin_left=15, x_margin_right=14, x_padding=33, y_margin_top=14, y_padding=16, y_margin_bottom=1)



        self.char_pos = Vector2(0, 0)
        self.main_character = AnimatedSprite("main"
                                             , Vector2(0, 0)
                                             , {"walkback": tuple(self.character_spritesheet.load_row_strip(8, colorkey=-1)[1:9]),
                                                "walkleft": tuple(self.character_spritesheet.load_row_strip(9, colorkey=-1)[1:9]),
                                                "walkdown": tuple(self.character_spritesheet.load_row_strip(10, colorkey=-1)[1:9]),
                                                "walkright": tuple(self.character_spritesheet.load_row_strip(11, colorkey=-1)[1:9]),
                                                "standright": tuple(self.character_spritesheet.load_row_strip(11, colorkey=-1)[0:0]),
                                                "standleft": tuple(self.character_spritesheet.load_row_strip(9, colorkey=-1)[0:0]),
                                                "standback": tuple(self.character_spritesheet.load_row_strip(8, colorkey=-1)[0:0]),
                                                "standdown": tuple(self.character_spritesheet.load_row_strip(10, colorkey=-1)[0:0]),
                                                "fall": tuple(self.character_spritesheet.load_row_strip(20, colorkey=-1)[:6])}
                                             , default_animation_key="walkdown",
                                             loop=True)

        self.scroll_handler = ScrollTileHandler(
            tile_size=self.tileset.size,
            screen_dims=self.screen.get_size(),
            focal_point=self.char_pos.as_tuple(),
            width_dim_p=WIDTH_DIM_P,
            height_dim_p=HEIGHT_DIM_P
        )
        print(self.chunk_renderer.tile_state((100, 100)))

    def head_left(self, delta_ms):
        self.head_direction(delta_ms, -DELTA_AMT_PER_100MS, 0)
        self.main_character.set_animation('walkleft')

    def head_right(self, delta_ms):
        self.head_direction(delta_ms, DELTA_AMT_PER_100MS, 0)
        self.main_character.set_animation('walkright')

    def head_up(self, delta_ms):
        self.head_direction(delta_ms, 0, -DELTA_AMT_PER_100MS)
        self.main_character.set_animation('walkback')

    def head_down(self, delta_ms):
        self.head_direction(delta_ms, 0, DELTA_AMT_PER_100MS)
        self.main_character.set_animation('walkdown')

    def head_direction(self, delta_ms, delta_x, delta_y):
        self.update_char_pos(delta_ms, delta_x, delta_y)
        _ = self.scroll_handler.update(focal_point=self.char_pos.as_tuple())

    def update_char_pos(self, delta_ms, delta_x, delta_y):
        og_pos = self.char_pos
        self.char_pos += Vector2(delta_x, delta_y) * round(delta_ms / 100, 1)
        char_tile = sec_util.sector_from_coord(
            coord=self.char_pos.as_tuple(),
            sec_dims=self.tileset.size)

        if self.chunk_renderer.tile_state(char_tile)[BG_TILE_STATE_HEADER] != 33:
            self.char_pos = og_pos

    def draw_tilemap(self):
        if self.tilemap and self.tilemap.rect:
            rect = pygame.Rect(self.tilemap.rect[0] - self.scroll_handler.WorldCoordOffsetScrollTileCoord[0],
                               self.tilemap.rect[1] - self.scroll_handler.WorldCoordOffsetScrollTileCoord[1],
                               self.tilemap.rect[2],
                               self.tilemap.rect[3])
            self.screen.blit(self.tilemap.image, rect)

    def draw_hud_text(self):
        help.draw_text(f"Scroll: {str((round(self.scroll_handler.Scroll[0], 2), round(self.scroll_handler.Scroll[1], 2)))}", self.screen)
        help.draw_text(f"WorldTile: {str(self.scroll_handler.WorldTile)}", self.screen, offset_rect=Rectangle.from_tuple((0, self.screen.get_height() - 25, 20, self.screen_width)))
        help.draw_text(f"PosCurrTile: {str(self.scroll_handler.PosInCurrentTile)}", self.screen,
                       offset_rect=Rectangle.from_tuple((0, self.screen.get_height() - 75, 20, self.screen_width)))
        help.draw_text(f"WorldPosOffsetScroll: {str(self.scroll_handler.WorldCoordOffsetScrollTileCoord)}", self.screen,
                       offset_rect=Rectangle.from_tuple((0, self.screen.get_height() - 100, 20, self.screen_width)))

    def draw_minimap(self):
        self.screen.blit(self.minimap_surface,
                         (self.screen.get_width() - self.minimap_surface.get_width(),
                          self.screen.get_height() - self.minimap_surface.get_height()))

    def draw_prizes(self):
        chunk_size = self.chunk_renderer.chunk_size
        pxl_size = self.tileset.size
        chunk_pxls = (chunk_size[0] * pxl_size[0], chunk_size[1] * pxl_size[1])

        prizes = self.chunk_renderer.visible_tile_idx_by_condition(
            scroll=self.scroll_handler.Scroll,
            window_size=self.screen.get_size(),
            tile_evaluator=lambda dic: dic.get(TILE_HAS_PRIZE_HEADER, False) == True,
            chunk_pxls=chunk_pxls,
            buffer_tiles=(1, 1))

        for idx, state in prizes.items():
            window_coord = self.scroll_handler.tile_window_coord(idx)
            help.draw_circle(surface=self.screen, center=Vector2.from_tuple(window_coord), radius=10, color=Color.YELLOW, width=5)

    def draw_character(self):
        self.main_character.set_pos(Vector2.from_tuple(self.scroll_handler.WindowCoords), SpriteOriginPosition.BOTTOM_MIDDLE)
        self.main_character.blit(self.screen,
                                 display_handle=lambda: self.debug_mode_toggle.value,
                                 display_rect=lambda: self.debug_mode_toggle.value)

    def draw(self, frames: int, debug_mode: bool = False):
        self.draw_tilemap()
        self.draw_prizes()
        # self.draw_hud_text()
        self.draw_minimap()
        self.draw_character()


    def update_tilemap(self):
        chunk_size = self.chunk_renderer.chunk_size
        pxl_size = self.tileset.size
        chunk_pxls = (chunk_size[0] * pxl_size[0], chunk_size[1] * pxl_size[1])

        tm_array = self.chunk_renderer.renderable_tile_type_array(scroll=self.scroll_handler.Scroll,
                                                                  window_size=self.screen.get_size(),
                                                                  key=BG_TILE_STATE_HEADER,
                                                                  chunk_pxls=chunk_pxls,
                                                                  buffer_tiles=(1, 1))

        self.tilemap = Tilemap(self.tileset,
                               map=tm_array,
                               surface_size=self.screen.get_size(),
                               buffer_tiles=(1, 1))



    def model_updater(self, delta_time_ms: int):
        # self.head_left(delta_time_ms)
        self.update_tilemap()
        self.update_minimap()


    def update_minimap(self):
        self.minimap_surface.fill(Color.BLACK.value)

        minimap_shape = self.minimap_surface.get_size()

        border = (5, 5)
        smallest_idx = (min(chunk[0] for chunk, data in self.chunk_renderer.chunk_data.items()), min(chunk[1] for chunk, data in self.chunk_renderer.chunk_data.items()))
        largest_idx = (max(chunk[0] for chunk, data in self.chunk_renderer.chunk_data.items()), max(chunk[1] for chunk, data in self.chunk_renderer.chunk_data.items()))



        width_space = minimap_shape[0]  - 2 * border[0]
        height_space = minimap_shape[1] - 2 * border[1]

        chunks_x = (largest_idx[1] - smallest_idx[1] + 1)
        chunks_y = (largest_idx[0] - smallest_idx[0] + 1)

        chunk_size_in_pxls_minimap = tuple([int(width_space / chunks_x),
                                           int(height_space / chunks_y)])

        player_tile = self.scroll_handler.WorldTile
        current_chunk = chunk_of_tile(player_tile, self.chunk_renderer.chunk_size)

        chunk_color_lookup = {
            (0, 0): Color.ORANGE,
        }
        if self.minimap_show_player_chunk_toggle.value:
            chunk_color_lookup[current_chunk] = Color.YELLOW

        for chunk, data in self.chunk_renderer.chunk_data.items():
            help.draw_box(surface=self.minimap_surface,
                          rect=Rectangle.from_tuple((
                              (chunk[1] - smallest_idx[1]) * chunk_size_in_pxls_minimap[0] + border[0],
                              (chunk[0] - smallest_idx[0]) * chunk_size_in_pxls_minimap[1] + border[1],
                              chunk_size_in_pxls_minimap[0],
                              chunk_size_in_pxls_minimap[1]
                          )),
                          color=chunk_color_lookup.get(chunk, Color.DARK_GRAY),
                          outline_color=Color.DIM_GRAY if self.minimap_show_chunk_outline_toggle.value else None
                          )

        # tile_color_lookup = {
        #     player_tile: Color.GREEN
        # }
        # if self.minimap_show_player_tile_toggle.value:
        #     for tile, data in self.chunk_renderer.chunk_data[current_chunk].items():
        #         help.draw_box(surface=self.minimap_surface,
        #                       rect=Rectangle(
        #                           x=(tile[1] - smallest_idx[1] * self.chunk_renderer.chunk_size[0]) * tile_size_in_pxls_minimap[0] + border[0],
        #                           y=(tile[0] - smallest_idx[0] * self.chunk_renderer.chunk_size[1]) * tile_size_in_pxls_minimap[1] + border[1],
        #                           width=tile_size_in_pxls_minimap[0],
        #                           height=tile_size_in_pxls_minimap[1]
        #                       ),
        #                       color=tile_color_lookup.get(tile, Color.DARK_GRAY),
        #                       outline_color=Color.DIM_GRAY if self.minimap_show_chunk_outline_toggle.value else None
        #                       )
        tilesize_in_minimap = (chunk_size_in_pxls_minimap[0] / self.chunk_renderer.chunk_size[0],
                               chunk_size_in_pxls_minimap[1] / self.chunk_renderer.chunk_size[1])
        smallest_tiles = (smallest_idx[0] * self.chunk_renderer.chunk_size[0], smallest_idx[1] * self.chunk_renderer.chunk_size[1])

        player_size_px = 10
        minimap_tile =((player_tile[0] - smallest_tiles[0]) * tilesize_in_minimap[1] + border[1],
                        (player_tile[1] - smallest_tiles[1]) * tilesize_in_minimap[0] + border[0])

        help.draw_circle(self.minimap_surface,
                         center=Vector2(minimap_tile[1] + player_size_px / 2, minimap_tile[0] + player_size_px / 2),
                         radius=player_size_px/2,
                         color=Color.BLUE,
                         outline_color=Color.WHITE)


    def sprite_updater(self, delta_time_ms: int):
        self.main_character.animate(delta_time_ms)

    def on_resize(self):
        pass

    def update_built_in_surfaces(self, surface_types: List[BuiltInSurfaceType]):
        pass


if __name__ == "__main__":

    game = ChunkGame()
    game.main()


from coopstructs.vectors import Vector2
from coopstructs.toggles import BooleanToggleable, EnumToggleable
from coopgame.gameTemplate import GameTemplate
from coopgraph.gridSelectPolicies import TogglePolicy
from coopgraph.toggles import BooleanToggleable
from cooptools.colors import Color
from coopgame.linedrawing.unitpathhandler import UnitPathHandler
import numpy as np
import coopgame.renderedObjectHandling.MatrixManipulation as mm
from coopgame.grids_graphs.pygridgraphHandler import PyGridGraphHandler
import coopgame.pygamehelpers as help
import pygame
from coopgame.spriteHandling.spritesheet import SpriteSheet
from coopgame.spriteHandling.sprites import AnimatedSprite, SpriteOriginPosition
from enum import Enum
import random as rnd

class Seasons(Enum):
    WINTER = 1
    SPRING = 2
    SUMMER = 3
    FALL = 4

class GridGraphGame(GameTemplate):

    def register_keys(self):
        self.register_action_to_keys((pygame.K_t,),
                                     lambda: self.gg_handler.show_toggled.visibility_toggle())
        self.register_action_to_keys((pygame.K_d,),
                                     lambda: self.gg_handler.allow_diag.visibility_toggle(on_toggle_callbacks=[lambda x: self.redraw_graph()]))
        self.register_action_to_keys((pygame.K_s,),
                                    lambda: self.season_toggle.toggle(on_toggle_callbacks=[lambda x: self.change_season(x)]))
        self.register_action_to_keys((pygame.K_g,),
                                    lambda: self.show_graph_toggle.toggle())

    def __init__(self):
        super().__init__(fullscreen=False)
        self.toggle_key = 'toggle'
        self.unit_path_handler = UnitPathHandler(self.toggle_key)
        self.toggle_bool_policy = TogglePolicy(key=self.toggle_key, toggle=BooleanToggleable(default=True))
        rows = 30
        cols = 20
        self.gg_handler = PyGridGraphHandler(rows, cols)

        self.x_rot = 1.1

        self.grid_surface = None
        self.graph_surface = None
        self.grid_overlay_surface = None

        self.last_hover_grid = None
        self.mouse_down_toggling_on = True

        self.time_since_report_stats = 0

        # Init Game
        self.gg_handler.show_toggled.visibility_toggle()
        self.redraw_graph_grid()


        # Init a tree
        self.season_toggle = EnumToggleable(Seasons)
        for ii in range(0, 5):
            self.init_tree(f"tree_{ii}", Vector2(rnd.randint(0, cols-1), rnd.randint(0, rows-1)))

        # toggles
        self.show_graph_toggle = BooleanToggleable(default=True)
        self.show_grid_toggle = BooleanToggleable(default=True)

    def init_tree(self, name: str, grid_pos: Vector2):
        tree_file = '../coopgame/assets/spriteSheets/489-4891674_tree-sprite-sheet-png.png'
        tree_spritesheet = SpriteSheet(tree_file, n_rows=4, n_columns=15, x_margin_left=17,
                                                 x_padding=34, y_margin_top=15, y_padding=17, y_margin_bottom=1)

        state = self.gg_handler.grid_graph.act_on_grid(grid_pos.y, grid_pos.x, [self.toggle_bool_policy])
        game_pos = self.gg_handler.grid.coord_from_grid_pos(grid_pos, self.game_area_rectangle)
        scaled = help.scaled_points([game_pos], self.get_grid_rotation_matrix())[0]
        tree_sprite = AnimatedSprite(id=name,
                                        init_pos=scaled,
                                        animation_dict={Seasons.WINTER: tuple(tree_spritesheet.load_row_strip(0, colorkey=-1)[:1]),
                                                            Seasons.SPRING: tuple(tree_spritesheet.load_row_strip(0, colorkey=-1)[:15]),
                                                            Seasons.SUMMER: tuple(tree_spritesheet.load_row_strip(1, colorkey=-1)[:15]),
                                                            Seasons.FALL: tuple(tree_spritesheet.load_row_strip(2, colorkey=-1)[:15])},
                                        default_animation_key=Seasons.WINTER,
                                        animation_cycle_ms=500,
                                        sprite_origin_position=SpriteOriginPosition.BOTTOM_MIDDLE,
                                        height=100,
                                        width=100
                                     )
        self.sprites[name] = tree_sprite

    def change_season(self, season):
        for id, sprite in self.sprites.items():
            sprite.set_animation(season)

    def redraw_graph_grid(self):
        self.grid_surface = help.init_surface(self.screen.get_size())
        self.graph_surface = help.init_surface(self.screen.get_size())
        self.grid_overlay_surface = help.init_surface(self.screen.get_size())

        self.gg_handler.redraw_graph_grid_overlay(self.grid_surface,
                                                  self.graph_surface,
                                                  self.grid_overlay_surface,
                                                  self.get_grid_rotation_matrix(),
                                                  predefined_highlights={Vector2(1, 1): Color.PURPLE,
                                                            Vector2(5, 5): Color.YELLOW})


    def redraw_grid_overlay(self):
        self.grid_overlay_surface = help.init_surface(self.screen.get_size())

        self.gg_handler.redraw_grid_overlay(self.grid_overlay_surface,
                                          self.get_grid_rotation_matrix(),
                                          predefined_highlights={Vector2(1, 1): Color.PURPLE,
                                                                 Vector2(5, 5): Color.YELLOW},)

    def redraw_graph(self):
        self.graph_surface = help.init_surface(self.screen.get_size())

        self.gg_handler.redraw_graph(self.graph_surface,
                                          self.get_grid_rotation_matrix(),)

    def draw(self, frames:int, debug_mode: bool = False):
        if self.show_grid_toggle.value:
            self.screen.blit(self.grid_surface, (0, 0))

        if self.show_graph_toggle.value:
            self.screen.blit(self.graph_surface, (0, 0))

        self.redraw_grid_overlay()
        self.screen.blit(self.grid_overlay_surface, (0, 0))

        # draw sprites
        self.sprite_drawer(self.screen)

    def handle_left_mouse_down(self, pressed_keys, mouse_pos_as_vector: Vector2):
        grid_pos = self.gg_handler.mouse_grid_pos(self.game_area_rectangle, help.mouse_pos_as_vector(), self.get_grid_rotation_matrix())

        if grid_pos is not None:
            state = self.gg_handler.grid_graph.act_on_grid(grid_pos.y, grid_pos.x, [self.toggle_bool_policy])
            if state[self.gg_handler.toggle_key] == True:
                self.mouse_down_toggling_on = True
            else:
                self.mouse_down_toggling_on = False

            self.redraw_graph()

    def handle_right_mouse_down(self, pressed_keys, mouse_pos_as_vector: Vector2):
        clicked_pos = self.gg_handler.mouse_grid_pos(self.game_area_rectangle, help.mouse_pos_as_vector(), self.get_grid_rotation_matrix())
        self.gg_handler.pff.set_astar_origin(clicked_pos)

    def need_to_toggle_at_pos(self, mouse_grid_pos):
        toggle_at_pos = self.gg_handler[mouse_grid_pos]
        if toggle_at_pos is None:
            return False

        return self.gg_handler[mouse_grid_pos].value != self.mouse_down_toggling_on

    def handle_hover_over(self, pressed_keys, mouse_pos_as_vector: Vector2):
        mouse_grid_pos = self.gg_handler.mouse_grid_pos(self.game_area_rectangle, help.mouse_pos_as_vector(), draw_scale_matrix=self.get_grid_rotation_matrix())

        if pressed_keys and mouse_grid_pos != self.last_hover_grid and self.need_to_toggle_at_pos(mouse_grid_pos):
            # handle drag across cells
            self.gg_handler.grid.act_on_grid(mouse_grid_pos.y, mouse_grid_pos.x, [self.gg_handler.toggle_bool_policy])

        self.last_hover_grid = mouse_grid_pos

        self.gg_handler.handle_hover_over(window_rect=self.game_area_rectangle,
                                          grid_rotation_matrix=self.get_grid_rotation_matrix())

    def get_grid_rotation_matrix(self):
        orientation = np.array([0, 0, 1])
        rotation_unit_vector = mm.rotation_unit_vector(orientation)
        matrixz = mm.rotateAroundAxis((self.screen_width / 2, self.screen_height / 2, 0), rotation_unit_vector,
                                      np.pi / 4)
        orientation = np.array([1, 0, 0])
        rotation_unit_vector = mm.rotation_unit_vector(orientation)
        matrixx = mm.rotateAroundAxis((self.screen_width / 2, self.screen_height / 2, 0), rotation_unit_vector,
                                      self.x_rot)
        matrix = matrixx.dot(matrixz)
        return matrix

if __name__ == "__main__":
    import logging

    logging.basicConfig(level=logging.INFO)
    logging.getLogger('coopgame').setLevel(logging.INFO)
    g = GridGraphGame()
    g.main()


from coopgame.gameTemplate import GameTemplate
import coopgame.pygamehelpers as help
from cooptools.colors import Color
from coopgame.grids_graphs.pygraphhandler import PyGraphHandler
from cooptools.toggles import BooleanToggleable
from coopgame.handleKeyPressedArgs import InputState, InputStateHandler, InputAction, InputEvent, InputEventType, basic_key_actions, CallbackPackage
from coopgame.pygame_k_constant_names import PyKeys, PyMouse
from coopgraph.graphs import Graph, Node,AStarResults, Edge
from typing import Tuple, List
import string
import random as rnd
import math

from coopgame.grids_graphs.pygraphhandler import GraphSurfaceType
import coopgame.grids_graphs.draw_graph_utils as g_util
import coopgame.linedrawing.line_draw_utils as line_u
import coopgame.pointdrawing.point_draw_utils as point_u
import coopgame.label_drawing.label_drawing_utils as lab_u


NODE_COLOR = Color.AQUA
ENABLED_EDGE_COLOR = Color.ORANGE
DISABLED_EDGE_COLOR = Color.MAROON
NODE_LABEL_COLOR = Color.WHEAT
EDGE_LABEL_COLOR = Color.WHEAT
ARTICULATION_POINTS_COLOR = Color.PURPLE
DIRECTIONALITY_INDICATOR_COLOR = Color.GREEN
NODE_RADIUS = 5

def large_circuit(x_bounds: Tuple[float, float],
                  y_bounds: Tuple[float, float],
                  spread: int = 100,
                  n_nodes: int = 20):
    alphabet=string.ascii_lowercase

    nodes = []
    rnd.seed(0)
    for ii in range(0, n_nodes):
        nodes.append(
            Node(name=alphabet[ii],
                 pos=((x_bounds[1] - x_bounds[0]) / 2 + x_bounds[0] + (x_bounds[1] - x_bounds[0]) * math.sin(ii / n_nodes * 2 * math.pi) + rnd.randint(-spread, spread),
                     (y_bounds[1] - y_bounds[0]) / 2 + y_bounds[0] + (y_bounds[1] - y_bounds[0])  * math.cos(ii / n_nodes * 2 * math.pi) + rnd.randint(-spread, spread))))

    graph_dict = {}

    for ii, node in enumerate(nodes):
        n_connections = rnd.randint(1, 4)

        for n in range(1, n_connections + 1):
            if ii + n >= n_nodes:
                index = ii + n - n_nodes
            else:
                index = ii + n
            graph_dict.setdefault(node, []).append(nodes[index])

    # add an orphan
    graph_dict[Node(name='Orphan_0', pos=(x_bounds[0] + (x_bounds[1] - x_bounds[0]) / 2, y_bounds[0] + (y_bounds[1] - y_bounds[0]) / 2))] = []

    # add a source
    graph_dict[Node(name='Source_0', pos=(x_bounds[0] + (x_bounds[1] - x_bounds[0]) / 2 - 50, y_bounds[0] + (y_bounds[1] - y_bounds[0]) / 2))] = [next(k for k, v in graph_dict.items())]

    # add a sink
    sink = Node(name='Sink_0', pos=(x_bounds[0] + (x_bounds[1] - x_bounds[0]) / 2 + 50, y_bounds[0] + (y_bounds[1] - y_bounds[0]) / 2))
    graph_dict[next(k for k, v in graph_dict.items())].append(sink)
    graph_dict[sink] = []

    return graph_dict


class GraphGame_Basic(GameTemplate):

    def __init__(self):


        self.graph_surface = None
        self.graph_overlay = None

        # toggles
        self.show_graph_base_toggle = BooleanToggleable(default=True)
        self.show_graph_overlay_toggle = BooleanToggleable(default=True)

        super().__init__(input_state_handler=self.define_input_handler())

        self.graph = Graph(graph_dict=large_circuit((0.25 * self.screen_width, 0.75 * self.screen_width),
                                                    (0.25 * self.screen_height, 0.75 * self.screen_height),
                                                    spread=25))
        self.graph_handler = PyGraphHandler(
            screen=self.screen,
            graph_getter=lambda: self.graph,
            route_draw_args_getter=lambda: [g_util.RouteDrawArgs(
               id='Route',
                route=self.get_a_route(),
                edge_draw_args=line_u.DrawLineArgs(
                    color=Color.TEAL,
                    width=3,
                    draw_label_args=lab_u.DrawLabelArgs(
                        color=Color.TEAL
                    ),
                    directionality_draw_args=line_u.DirectionalityIndicatorDrawArgs(
                        color=Color.TEAL
                    )
                ),
                node_draw_args=point_u.DrawPointArgs(
                    color=Color.TEAL,
                    radius=5
                ),
                node_label_args=lab_u.DrawLabelArgs(
                    color=Color.TEAL
                )
            )]
        )

    def on_resize(self):
        self.graph_handler.redraw()

    def define_input_handler(self):
        return InputStateHandler(
            quit_callback_package=CallbackPackage(down=lambda x: self.quit()),
            debug_callback_package=CallbackPackage(down=lambda x: self.debug_mode_toggle.toggle()),
            fullscreen_callback_package=CallbackPackage(down=lambda x: self.toggle_fullscreen()),
            key_actions=[
                InputAction([InputEvent(PyKeys.K_r, event_type=InputEventType.DOWN)], callback=lambda x: self.redraw()),
                InputAction([InputEvent(PyKeys.K_1, event_type=InputEventType.DOWN)], callback=lambda x: self.graph_handler.toggle_surface([GraphSurfaceType.EDGES_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_2, event_type=InputEventType.DOWN)], callback=lambda x: self.graph_handler.toggle_surface([GraphSurfaceType.NODES_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_3, event_type=InputEventType.DOWN)], callback=lambda x: self.graph_handler.toggle_surface([GraphSurfaceType.EDGE_LABELS_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_4, event_type=InputEventType.DOWN)], callback=lambda x: self.graph_handler.toggle_surface([GraphSurfaceType.NODE_LABELS_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_5, event_type=InputEventType.DOWN)], callback=lambda x: self.graph_handler.toggle_surface([GraphSurfaceType.OVERLAY_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_6, event_type=InputEventType.DOWN)], callback=lambda x: self.graph_handler.toggle_surface([GraphSurfaceType.ROUTES_SURFACE_ID])),
                InputAction([InputEvent(PyKeys.K_KP_PLUS, event_type=InputEventType.DOWN)], callback=lambda x: self.graph_handler.show_all()),
                InputAction([InputEvent(PyKeys.K_KP_MINUS, event_type=InputEventType.DOWN)], callback=lambda x: self.graph_handler.hide_all()),

        ])

    def redraw(self):
        self.graph_handler.redraw()

    def draw(self, frames: int, debug_mode: bool = False):
        self.graph_handler.render(self.screen)

    def get_a_route(self) -> List[Edge]:
        s = rnd.choice(self.graph.nodes)
        e = rnd.choice(self.graph.nodes)

        if s == e \
                or s in self.graph.Sinks \
                or e in self.graph.Sources \
                or s in self.graph.Orphans\
                or e in self.graph.Orphans:
            return []

        waypoints = self.graph.astar(
            start=s,
            end=e
        )

        route = [self.graph.edge_between(waypoints.path[ii - 1], waypoints.path[ii]) for ii in range(1, len(waypoints.path))]

        return route


if __name__ == "__main__":
    g = GraphGame_Basic()
    g.main()


from cooptools.toggles import EnumToggleable
from enum import Enum
from coopgame.models.primitives.gameObject import GameObject
from typing import List
from cooptools.geometry_utils import vector_utils as vec

class TreePhase(Enum):
    SEED = 0
    SAPLING = 1
    YOUTH = 2
    ADULT = 3
    ANCIENT = 4

class TreeType:
    OAK = 0
    WILLOW = 1
    MAPLE = 2
    BIRCH = 3


class Tree(GameObject):
    def __init__(self,
                 id,
                 pos: vec.FloatVec,
                 type: TreeType,
                 age: int = None,
                 tags: List[str] = None):
        super().__init__(name=id, pos=pos, tags=tags)
        self.id = id
        self.type = type
        self._age_phase = EnumToggleable(TreePhase, TreePhase.SEED)
        self._age = age if age else 0


